import unittest
import pyrtma
import pyrtma.core_defs as cd
import sys
from .test_msg_defs import test_defs as td

sys.path.append("..")
from examples.msg_defs import example_messages as em


class TestMessageVersion(unittest.TestCase):
    """Test that the core defs and example/test defs were built
    with the same version of pyrtma

    It is easy to miss this step before deployment, so this test will catch it
    """

    def test_version(self):
        with self.subTest("core def version"):
            self.assertEqual(pyrtma.__version__, cd.COMPILED_PYRTMA_VERSION)
        with self.subTest("test def version"):
            self.assertEqual(pyrtma.__version__, td.COMPILED_PYRTMA_VERSION)
        with self.subTest("example def version"):
            self.assertEqual(pyrtma.__version__, em.COMPILED_PYRTMA_VERSION)



import unittest
import random
import threading
import time
import logging

import pyrtma
from .test_msg_defs import test_defs as td
import time
from pyrtma.client import client_context
from pyrtma.message import *
import pyrtma.manager
from pyrtma.manager import MessageManager


class TestSync(unittest.TestCase):
    """Test sending messages through MessageManager."""

    def setUp(self):
        self.port = random.randint(1000, 10000)  # random port
        self.addr = f"127.0.0.1:{self.port}"

        self.manager = MessageManager(
            ip_address="127.0.0.1",
            port=self.port,
            timecode=False,
            log_level=logging.ERROR,
            debug=False,
            send_msg_timing=True,
        )
        self.manager_thread = threading.Thread(
            target=self.manager.run,
        )
        self.manager_thread.start()
        time.sleep(0.250)

    def tearDown(self):
        self.manager.close()
        self.manager_thread.join()

    def test_version_mismatch(self):
        with client_context(server_name=self.addr) as publisher:
            with client_context(
                server_name=self.addr, msg_list=[td.MT_SET_START]
            ) as subscriber:
                time.sleep(0.250)

                header = MessageHeader()
                header.msg_type = td.MT_SET_START
                header.msg_count = 1
                header.send_time = time.perf_counter()
                header.num_data_bytes = 0

                # Change the version to create a mismatch between sender and receiver
                header.version = 0xDEADBEEF

                publisher.forward_message(header, td.MDF_SET_START())

                with self.assertRaises(pyrtma.message.InvalidMessageDefinition):
                    msg = subscriber.read_message(timeout=0.100, sync_check=True)

    def test_size_mismatch(self):
        with client_context(server_name=self.addr) as publisher:
            with client_context(
                server_name=self.addr, msg_list=[td.MT_SET_START]
            ) as subscriber:
                time.sleep(0.250)

                header = MessageHeader()
                header.msg_type = td.MT_SET_START
                header.msg_count = 1
                header.send_time = time.perf_counter()

                # Create a size mismatch between sender and receiver
                data = td.MDF_TRIAL_METADATA()
                header.num_data_bytes = data.type_size

                publisher.forward_message(header, data)

                with self.assertRaises(pyrtma.message.InvalidMessageDefinition):
                    msg = subscriber.read_message(timeout=0.100, sync_check=True)

        time.sleep(0.5)



import ctypes
import random
import string
import decimal
from typing import List, TypeVar


def _random_str(length: int) -> str:
    """Utility to generate random string

    Args:
        length (int): String length

    Returns:
        str: Random string
    """
    if length > 1:
        length -= 1

    return "".join(random.choice(string.printable) for _ in range(length))


def _random_int_array(length: int, min: int = 0, max: int = 9) -> List[int]:
    """Utility to generate random int array

    Args:
        length (int): Array length
        min (int, optional): Minimum value. Defaults to 0.
        max (int, optional): Maximum value. Defaults to 9.

    Returns:
        List[int]: List of random ints
    """
    return [random.randint(min, max) for _ in range(length)]


def _random_float_array(length: int) -> List[float]:
    """Utility to generate random float array

    Restricted to values that can be represented by 32-bit float

    Args:
        length (int): Array length

    Returns:
        List[float]: List of random floats
    """
    context = decimal.Context(prec=5, rounding=decimal.ROUND_DOWN)
    return [round(random.random(), 6) for _ in range(length)]


def _random_double_array(length: int) -> List[float]:
    """Utility to generate random float array

    Values are of double (64-bit) precision

    Args:
        length (int): Array length

    Returns:
        List[float]: List of random floats
    """
    return [random.random() for _ in range(length)]


def _random_byte_array(length: int) -> bytes:
    """Utility to generate random byte string

    Args:
        length (int): byte string length

    Returns:
        bytes: Random byte string
    """
    return bytes([random.randint(0, 255) for _ in range(length)])


CStruct = TypeVar("CStruct", bound=ctypes.Structure)


def _random_struct(obj: CStruct) -> CStruct:
    """Utility to generate random CStructure values

    Args:
        obj (CStruct): CStructure instance

    Returns:
        CStruct: CStructure with random values
    """
    for _name, ftype, *_ in obj._fields_:
        name = _name[1:] if _name[0] == "_" else _name
        if issubclass(ftype, ctypes.Structure):
            setattr(obj, name, _random_struct(getattr(obj, name)))
        elif issubclass(ftype, ctypes.Array):
            length: int = ftype._length_  # type: ignore
            etype: type = ftype._type_  # type: ignore
            if issubclass(etype, ctypes.Structure):
                for i in range(length):
                    getattr(obj, name)[i] = _random_struct(getattr(obj, name)[i])
            elif etype is ctypes.c_char:
                setattr(obj, name, _random_str(length))
            elif etype is ctypes.c_ubyte:
                getattr(obj, name)[:] = _random_byte_array(length)
            elif etype is ctypes.c_int8:
                getattr(obj, name)[:] = _random_int_array(
                    length, min=-(2**7), max=2**7 - 1
                )
            elif etype is ctypes.c_uint8:
                getattr(obj, name)[:] = _random_int_array(length, min=0, max=2**8 - 1)
            elif etype in (ctypes.c_short, ctypes.c_int16):
                getattr(obj, name)[:] = _random_int_array(
                    length, min=-(2**15), max=2**15 - 1
                )
            elif etype in (ctypes.c_ushort, ctypes.c_uint16):
                getattr(obj, name)[:] = _random_int_array(length, min=0, max=2**16 - 1)
            elif etype in (ctypes.c_int, ctypes.c_long, ctypes.c_int32):
                getattr(obj, name)[:] = _random_int_array(
                    length, min=-(2**31), max=2**31 - 1
                )
            elif etype in (ctypes.c_uint, ctypes.c_ulong, ctypes.c_uint32):
                getattr(obj, name)[:] = _random_int_array(length, min=0, max=2**32 - 1)
            elif etype in (ctypes.c_longlong, ctypes.c_int64):
                getattr(obj, name)[:] = _random_int_array(
                    length, min=-(2**63), max=2**63 - 1
                )
            elif etype in (ctypes.c_ulonglong, ctypes.c_uint64):
                getattr(obj, name)[:] = _random_int_array(length, min=0, max=2**64 - 1)
            elif etype is ctypes.c_float:
                getattr(obj, name)[:] = _random_float_array(length)
            elif etype is ctypes.c_double:
                getattr(obj, name)[:] = _random_double_array(length)
        elif ftype is ctypes.c_char:
            setattr(obj, name, _random_str(1))
        elif ftype is ctypes.c_ubyte:
            setattr(obj, name, random.randint(0, 255))
        elif ftype is ctypes.c_int8:
            setattr(obj, name, random.randint(-(2**7), 2**7 - 1))
        elif ftype is ctypes.c_uint8:
            setattr(obj, name, random.randint(0, 2**8 - 1))
        elif ftype in (ctypes.c_short, ctypes.c_int16):
            setattr(obj, name, random.randint(-(2**15), 2**15 - 1))
        elif ftype in (ctypes.c_ushort, ctypes.c_uint16):
            setattr(obj, name, random.randint(0, 2**16 - 1))
        elif ftype in (ctypes.c_int, ctypes.c_long, ctypes.c_int32):
            setattr(obj, name, random.randint(-(2**31), 2**31 - 1))
        elif ftype in (ctypes.c_uint, ctypes.c_ulong, ctypes.c_uint32):
            setattr(obj, name, random.randint(0, 2**32 - 1))
        elif ftype in (ctypes.c_longlong, ctypes.c_int64):
            setattr(obj, name, random.randint(-(2**63), 2**63 - 1))
        elif ftype in (ctypes.c_ulonglong, ctypes.c_uint64):
            setattr(obj, name, random.randint(0, 2**64 - 1))
        elif ftype is ctypes.c_float:
            setattr(obj, name, random.random())
        elif ftype is ctypes.c_double:
            setattr(obj, name, random.random())

    return obj



import pyrtma
import os
import sys
import shutil
import tempfile

from pyrtma.utils.quicklogger_reader import QLFileHeader
from ..data_formatter import DataFormatter
from typing import ClassVar, IO, List


class QLFormatter(DataFormatter):
    name: ClassVar[str] = "quicklogger"
    mode: ClassVar[str] = "wb"
    ext: ClassVar[str] = ".bin"

    def __init__(self, fd: IO[bytes]):
        # Initialize the quicklogger file header
        hdr = pyrtma.get_header_cls()
        self.ql_header = QLFileHeader()
        self.ql_header.format_version = 1
        self.ql_header.num_messages = 0
        self.ql_header.message_header_size = hdr().size
        self.ql_header.data_block_offset_size = 4
        self.ql_header.total_bytes = self.ql_header.size

        self.ofs = 0
        self.offsets: list[int] = []
        self.num_writes = 0

        self.data_tmp = tempfile.NamedTemporaryFile()
        # print(f"TempFile: {self.data_tmp.name}")

        super().__init__(fd)

    def format_header(self) -> bytes:
        # Initially add a placeholder for the
        return bytes(self.ql_header)

    def format_message(self, msg: pyrtma.Message) -> bytes:
        self.offsets.append(self.ofs)
        self.ofs += msg.data.type_size

        self.ql_header.num_messages += 1
        self.ql_header.num_data_bytes += msg.data.type_size
        self.ql_header.total_bytes += (
            self.ql_header.data_block_offset_size + msg.data.type_size + msg.header.size
        )

        return bytes(msg.header)

    def write(self, wbuf: List[pyrtma.Message]):
        super().write(wbuf)
        self.update_file_header()

        # Append the message data to the temp file
        # Note: This data gets copied and appended at the end
        # to preserve the QL v1 format
        self.data_tmp.writelines(bytes(msg.data) for msg in wbuf)
        self.num_writes += 1

    def write_offsets(self):
        ofs = b"".join(x.to_bytes(4, byteorder=sys.byteorder) for x in self.offsets)
        self.fd.write(ofs)

    def update_file_header(self):
        self.fd.seek(0, os.SEEK_SET)
        self.fd.write(bytes(self.ql_header))
        self.fd.seek(0, os.SEEK_END)

    def copy_data(self):
        # Reset the file pointer back to the start
        self.data_tmp.seek(os.SEEK_SET)

        # Append the message data tempfile to the data set file
        shutil.copyfileobj(self.data_tmp, self.fd)

    def finalize(self, wbuf: List[pyrtma.Message]):
        if self.num_writes > 0:
            self.write(wbuf)
            self.write_offsets()
            self.copy_data()
        else:
            # Skip copying data to tmp file if
            # our collection only needs one write
            super().write(wbuf)
            self.update_file_header()
            self.write_offsets()
            self.fd.writelines(bytes(msg.data) for msg in wbuf)

        self.data_tmp.close()



import time

import pyrtma
import pyrtma.core_defs as cd

from pyrtma.exceptions import UnknownMessageType
from typing import Optional, cast

from .metadata import LoggingMetadata
from .data_collection import DataCollection
from .data_set import DataSet
from .data_formatter import get_formatter
from .exceptions import *


class DataLogger:
    def __init__(self, rtma_server_ip: str, log_level: int):
        self.mod = pyrtma.Client(module_id=cd.MID_DATA_LOGGER, name="data_logger")
        self.mod.logger.level = log_level
        self.mod.connect(rtma_server_ip, logger_status=True)
        self.ctrl_msg_types = [
            cd.MT_DATA_LOGGER_START,
            cd.MT_DATA_LOGGER_STOP,
            cd.MT_DATA_LOGGER_PAUSE,
            cd.MT_DATA_LOGGER_RESUME,
            cd.MT_DATA_LOGGER_RESET,
            cd.MT_ADD_DATA_COLLECTION,
            cd.MT_ADD_DATA_SET,
            cd.MT_REMOVE_DATA_COLLECTION,
            cd.MT_REMOVE_DATA_SET,
            cd.MT_DATA_COLLECTION_CONFIG_REQUEST,
            cd.MT_DATA_LOGGER_STATUS_REQUEST,
            cd.MT_DATA_LOGGER_METADATA_UPDATE,
            cd.MT_DATA_LOGGER_METADATA_REQUEST,
            cd.MT_EXIT,
        ]

        self.mod.subscribe(self.ctrl_msg_types)

        self.mod.send_module_ready()

        self.mod.info("DataLogger connected and waiting for configuration.")

        self.metadata = LoggingMetadata()
        self.collection: Optional[DataCollection] = None

        self._running = False
        self._recording = False
        self._paused = False
        self._elapsed_time = 0.0

    def send_status(self):
        msg = cd.MDF_DATA_LOGGER_STATUS()
        msg.timestamp = time.time()
        if self._recording and self.collection:
            msg.elapsed_time = self.collection.elapsed_time
        else:
            msg.elapsed_time = -1
        msg.is_recording = self._recording
        msg.is_paused = self._paused

        self.mod.send_message(msg)

    def collection_info(self) -> cd.DATA_COLLECTION_INFO:
        if self.collection:
            c = cd.DATA_COLLECTION_INFO()

            c.name = self.collection.name
            c.save_path = str(self.collection.save_path)
            c.num_data_sets = len(self.collection.datasets)

            for i, ds in enumerate(self.collection.datasets):
                c.data_sets[i].name = ds.name
                c.data_sets[i].formatter = ds.formatter_cls.name
                c.data_sets[i].save_path = str(ds.file_path)
                c.data_sets[i].subdivide_interval = (
                    0
                    if isinstance(ds.subdivide_interval, float)
                    else ds.subdivide_interval
                )
                c.data_sets[i].msg_types[: len(ds.msg_types)] = ds.msg_types

        return c

    def send_config(self):
        msg = cd.MDF_DATA_COLLECTION_CONFIG()

        if self.collection is None:
            self.mod.send_message(msg)
            return

        msg.collection.name = self.collection.name
        msg.collection.base_path = str(self.collection.base_path)
        msg.collection.dir_fmt = self.collection.dir_fmt
        msg.collection.num_data_sets = len(self.collection.datasets)

        for i, ds in enumerate(self.collection.datasets):
            d = msg.collection.data_sets[i]

            d.name = ds.name
            d.sub_dir_fmt = ds.sub_dir_fmt
            d.file_name_fmt = ds.file_name_fmt
            d.formatter = ds.formatter_cls.name
            d.msg_types[: len(ds.msg_types)] = ds.msg_types
            d.subdivide_interval = (
                0 if isinstance(ds.subdivide_interval, float) else ds.subdivide_interval
            )

        self.mod.send_message(msg)

    def start_logging(self):
        if self.collection:
            if self._recording:
                raise DataCollectionInProgress(
                    "Cannot start collection while recording in progresss."
                )

            self.mod.unsubscribe(self.ctrl_msg_types)
            self.mod.subscribe([cd.ALL_MESSAGE_TYPES])
            self.collection.start()

            msg = cd.MDF_DATA_COLLECTION_STARTED()
            msg.collection = self.collection_info()
            self.mod.send_message(msg)

            self.mod.info("Data logging started.")
            self._recording = True
            self._paused = False
        else:
            self.mod.unsubscribe([cd.ALL_MESSAGE_TYPES])
            self.mod.subscribe(self.ctrl_msg_types)
            raise DataCollectionNotConfigured("Ignoring start logging request.")

        self.send_status()

    def stop_logging(self):
        if self.collection:
            if self._recording:
                self.mod.unsubscribe([cd.ALL_MESSAGE_TYPES])
                self.collection.stop()
                self.mod.subscribe(self.ctrl_msg_types)

                msg = cd.MDF_DATA_COLLECTION_STOPPED()
                msg.collection = self.collection_info()
                self.mod.send_message(msg)
                self.mod.send_signal(cd.MT_DATA_COLLECTION_SAVED)

                self.mod.info("Data logging stopped.")
            else:
                self.mod.warn("Logger not recording. Stop ignored.")
        else:
            raise DataCollectionNotConfigured("Ignoring stop logging request.")

        self._recording = False
        self._paused = False
        self.send_status()

    def pause_logging(self):
        if self.collection:
            self.collection.pause()
            self.mod.info("Data logging paused.")
            self._paused = True
        else:
            raise DataCollectionNotConfigured("Ignoring pause logging request.")

    def resume_logging(self):
        if self.collection:
            self.collection.resume()
            self.mod.info("Data logging resumed.")
            self._paused = False
        else:
            raise DataCollectionNotConfigured("Ignoring resume logging request.")

    def add_data_collection(self, msg: cd.MDF_ADD_DATA_COLLECTION):
        if self._recording:
            raise DataCollectionInProgress(
                "Cannot add data collection while recording in progresss."
            )

        if self.collection:
            self.mod.info(f"Closing previous collection: {self.collection.name}")
            self.collection.close()
            self.rm_data_collection()

        self.collection = DataCollection(
            msg.collection.name,
            msg.collection.base_path,
            msg.collection.dir_fmt,
            self.metadata,
        )
        self.mod.info(f"Created data collection: {msg.collection.name}")

        for i in range(msg.collection.num_data_sets):
            ds = msg.collection.data_sets[i]

            try:
                self.collection.add_data_set(
                    DataSet(
                        collection_name=self.collection.name,
                        name=ds.name,
                        sub_dir_fmt=ds.sub_dir_fmt,
                        file_name_fmt=ds.file_name_fmt,
                        msg_types=ds.msg_types[:],
                        formatter_cls=get_formatter(ds.formatter),
                        subdivide_interval=ds.subdivide_interval,
                        metadata=self.collection.metadata,
                    )
                )
            except (DataFormatterKeyError, InvalidFormatter, DataSetExistsError) as e:
                self.mod.error(e.args[0])
                self.send_error(f"{e.__class__.__name__}:{e.args[0]}")
                continue

    def add_data_set(self, msg: cd.MDF_ADD_DATA_SET):
        if self._recording:
            raise DataCollectionInProgress(
                "Cannot add data set while recording in progresss."
            )

        if self.collection is None:
            raise DataCollectionNotConfigured(
                "Cannot add data set without setting up a collection first."
            )

        if msg.collection_name != self.collection.name:
            DataCollectionNotFound(
                f"add_data_set: trying to add to an unknown collection: {msg.collection_name}"
            )

        ds = DataSet(
            collection_name=self.collection.name,
            name=msg.data_set.name,
            sub_dir_fmt=msg.data_set.sub_dir_fmt,
            file_name_fmt=msg.data_set.file_name_fmt,
            msg_types=msg.data_set.msg_types[:],
            formatter_cls=get_formatter(msg.data_set.formatter),
            subdivide_interval=msg.data_set.subdivide_interval,
            metadata=self.collection.metadata,
        )

        self.collection.add_data_set(ds)

    def rm_data_collection(self):
        if self._recording:
            raise DataCollectionInProgress(
                "Cannot remove collection while recording in progress."
            )

        if self.collection:
            self.mod.info(f"Removed data collection: {self.collection.name}")
            self.collection.close()
            self.collection = None

    def rm_data_set(self, msg: cd.MDF_REMOVE_DATA_SET):
        if self._recording:
            raise DataCollectionInProgress(
                "Cannot remove data set while recording in progress."
            )

        if self.collection:
            self.collection.rm_data_set(msg.name)

    def reset(self):
        if (c := self.collection) is not None:
            if self._recording:
                self.mod.warn("Reset while recording in progress. Stopping collection!")
                self.stop_logging()

            self.rm_data_collection()

        self.metadata.clear()
        self._recording = False
        self._paused = False
        self.mod.info(f"Reset DataLogger. All collection/sets are cleared")

    def update_metadata(self, json_str: str):
        if self._recording:
            raise DataCollectionInProgress(
                "Cannot update metadata while recording in progress."
            )

        self.metadata.update(json_str)

    def send_error(self, msg: str):
        err = cd.MDF_DATA_LOGGER_ERROR()
        err.msg = msg
        self.mod.send_message(err)

    def run(self):
        try:
            self._running = True
            while self._running:
                try:
                    msg = self.mod.read_message(0.100)
                except UnknownMessageType as e:
                    self.mod.warn(f"UnknownMessageType: {e.args[0]}")
                    continue

                if self.collection and self._recording:
                    self.collection.update(msg)

                if msg is None:
                    continue

                try:
                    if msg.type_id == cd.MT_DATA_LOGGER_START:
                        self.start_logging()
                    elif msg.type_id == cd.MT_DATA_LOGGER_STOP:
                        self.stop_logging()
                    elif msg.type_id == cd.MT_DATA_LOGGER_PAUSE:
                        self.pause_logging()
                    elif msg.type_id == cd.MT_DATA_LOGGER_RESUME:
                        self.resume_logging()
                    elif msg.type_id == cd.MT_ADD_DATA_COLLECTION:
                        self.add_data_collection(
                            cast(cd.MDF_ADD_DATA_COLLECTION, msg.data)
                        )
                    elif msg.type_id == cd.MT_ADD_DATA_SET:
                        self.add_data_set(cast(cd.MDF_ADD_DATA_SET, msg.data))
                    elif msg.type_id == cd.MT_REMOVE_DATA_COLLECTION:
                        self.rm_data_collection()
                    elif msg.type_id == cd.MT_REMOVE_DATA_SET:
                        self.rm_data_set(cast(cd.MDF_REMOVE_DATA_SET, msg.data))
                    elif msg.type_id == cd.MT_DATA_LOGGER_RESET:
                        self.reset()
                    elif msg.type_id == cd.MT_DATA_LOGGER_STATUS_REQUEST:
                        self.send_status()
                    elif msg.type_id == cd.MT_DATA_COLLECTION_CONFIG_REQUEST:
                        self.send_config()
                    elif msg.type_id == cd.MT_DATA_LOGGER_METADATA_UPDATE:
                        self.update_metadata(
                            cast(cd.MDF_DATA_LOGGER_METADATA_UPDATE, msg.data).json
                        )
                    elif msg.type_id == cd.MT_DATA_LOGGER_METADATA_REQUEST:
                        meta = cd.MDF_DATA_LOGGER_METADATA()
                        meta.json = self.metadata.to_json()
                        self.mod.send_message(meta)
                    elif msg.type_id == cd.MT_EXIT:
                        if msg.header.dest_mod_id == self.mod.module_id:
                            self._running = False
                            self.mod.info("Received EXIT request. Closing...")
                except DataLoggerError as e:
                    self.mod.error(e.args[0])
                    self.send_error(f"{e.__class__.__name__}:{e.args[0]}")
                    if self._recording:
                        self.stop_logging()

        except KeyboardInterrupt:
            pass
        finally:
            if self.collection and self._recording:
                self.stop_logging()

            self._running = False

            if self.collection:
                self.collection.close()

            if self.mod.connected:
                self.mod.disconnect()
            self.mod.info("DataLogger is exiting...")



import unittest
import random
import threading
import time
import logging

import pyrtma
import pyrtma.message
import pyrtma.manager
import pyrtma.context

from pyrtma.client import client_context
from pyrtma.manager import MessageManager

# Import message defs to add to pyrtma.msg_defs map
from .test_msg_defs.test_defs import *


class TestEncoding(unittest.TestCase):
    """Test sending messages through MessageManager."""

    def setUp(self):
        self.port = random.randint(1000, 10000)  # random port
        self.addr = f"127.0.0.1:{self.port}"

        self.manager = MessageManager(
            ip_address="127.0.0.1",
            port=self.port,
            timecode=False,
            debug=False,
            log_level=logging.ERROR,
            send_msg_timing=True,
        )
        self.manager_thread = threading.Thread(
            target=self.manager.run,
        )
        self.manager_thread.start()
        time.sleep(0.250)

    def tearDown(self):
        self.manager.close()
        self.manager_thread.join()

    def test_message_encoding(self):
        with client_context(server_name=self.addr) as publisher:
            with client_context(server_name=self.addr) as subscriber:
                time.sleep(0.250)

                ctx = pyrtma.context.get_context()
                for mdf in ctx.MDF.values():
                    if mdf.type_id > 1000:
                        # Subscribe from message type
                        with subscriber.subscription_context([mdf.type_id]):
                            time.sleep(0.01)

                            # Publish a random message of the subscribed type
                            in_msg = mdf.from_random()
                            publisher.send_message(in_msg)

                            # Wait for the message and compare for equality of the data segment
                            out_msg = subscriber.read_message(timeout=0.020)
                            if out_msg is None:
                                self.fail("Subscriber did not receive packet.")
                            else:
                                self.assertEqual(in_msg, out_msg.data)

        time.sleep(0.5)



"""pyrtma.client module

Includes :py:class:`~Client` class and associated exception classes
"""

import socket
import select
import time
import os
import ctypes
import logging

from contextlib import contextmanager

from .context import get_context
from .message import Message, get_msg_cls
from .message_data import MessageData
from .header import MessageHeader, get_header_cls
from .core_defs import ALL_MESSAGE_TYPES
from .validators import disable_message_validation
from .client_logging import RTMALogger, ClientLike
from .exceptions import (
    InvalidMessageDefinition,
    UnknownMessageType,
    ClientError,
    SocketOptionError,
    MessageManagerNotFound,
    NotConnectedError,
    ConnectionLost,
    AcknowledgementTimeout,
    InvalidDestinationModule,
    InvalidDestinationHost,
    InvalidSubscription,
)
from . import core_defs as cd

from functools import wraps
from typing import (
    Dict,
    Optional,
    Tuple,
    Type,
    List,
    Union,
    Iterable,
    Set,
    Callable,
    Any,
    TypeVar,
    cast,
    Union,
)
from warnings import warn

__all__ = [
    "ClientError",
    "MessageManagerNotFound",
    "NotConnectedError",
    "ConnectionLost",
    "AcknowledgementTimeout",
    "InvalidDestinationModule",
    "InvalidDestinationHost",
    "Client",
    "client_context",
]

F = TypeVar("F", bound=Callable[..., Any])


def requires_connection(func: F) -> F:
    """Decorator wrapper for Client methods that require a connection"""

    @wraps(func)
    def wrapper(self, *args, **kwargs) -> F:
        if not self.connected:
            raise NotConnectedError
        else:
            return func(self, *args, **kwargs)

    return cast(F, wrapper)


class Client(ClientLike):
    """RTMA Client interface

    Args:
        module_id (optional): Static module ID, which must be unique.
            Defaults to 0, which generates a dynamic module ID.
        host_id (optional): Host ID. Defaults to 0.
        timecode (optional): Add additional timecode fields to message
            header, used by some projects at RNEL. Defaults to False.
    """

    def __init__(
        self,
        module_id: int = 0,
        host_id: int = 0,
        timecode: bool = False,
        name: str = "",
    ):
        if module_id >= cd.DYN_MOD_ID_START or module_id < 0:
            raise ValueError(f"Module ID must be >= 0 and < {cd.DYN_MOD_ID_START}")

        self._module_id = module_id
        self._host_id = host_id
        self._msg_count = 0
        self._server = ("", -1)
        self._connected = False
        self._header_cls = get_header_cls(timecode)
        self._recv_buffer = bytearray(1024**2)
        self._sub_all = False
        self._subscribed_types: Set[int] = set()
        self._paused_types: Set[int] = set()
        self._dynamic_id: bool = module_id == 0
        self._sock = socket.socket()

        # Auto-assign a name if module-id is defined
        ctx = get_context()
        if name == "" and module_id != 0:
            for k, v in ctx.MID.items():
                if v == module_id:
                    self._name = k
                    break
            else:
                self._name = name
        else:
            self._name = name

        self._logger = RTMALogger(
            self._name or f"Module {module_id}", self, logging.INFO
        )
        # alias logger methods
        self.debug = self._logger.debug
        self.info = self._logger.info
        self.warning = self._logger.warning
        self.warn = self._logger.warn
        self.error = self._logger.error
        self.exception = self._logger.exception
        self.critical = self._logger.critical

    @property
    def logger(self) -> RTMALogger:
        return self._logger

    def __del__(self):
        if self._connected:
            try:
                self.disconnect()
            except Exception:
                """Silently ignore any errors at this point."""
                pass

    def _socket_connect(self, server_name: str):
        # Close the previously used socket
        self._connected = False
        self._sock.close()

        # Get the server ip info
        addr, port = server_name.split(":")
        self._server = (addr, int(port))

        # Create the tcp socket
        self._sock = socket.socket(
            family=socket.AF_INET, type=socket.SOCK_STREAM, proto=socket.IPPROTO_TCP
        )

        # Connect to the message server
        try:
            self._sock.connect(self._server)
            self._connected = True
        except ConnectionRefusedError as e:
            self._connected = False
            raise MessageManagerNotFound(
                f"No message manager server responding at {self.ip_addr}:{self.port}"
            ) from e

        # Disable Nagle Algorithm
        try:
            self._sock.setsockopt(socket.getprotobyname("tcp"), socket.TCP_NODELAY, 1)
            self._sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        except Exception as e:
            self._connected = False
            self._sock.close()
            raise SocketOptionError from e

    def _connect_helper(
        self, logger_status: bool, daemon_status: bool, allow_multiple: bool
    ) -> Message:
        """Called internally after _socket_connect

        Note that this is also used by web_manager
        """

        # Reset the module_id to zero for dynamic assignment
        if self._dynamic_id:
            self._module_id = 0

        msg = cd.MDF_CONNECT()
        msg.logger_status = int(logger_status)
        msg.daemon_status = int(daemon_status)

        msg2 = cd.MDF_CONNECT_V2()
        msg2.logger_status = int(logger_status)
        msg2.daemon_status = int(daemon_status)
        msg2.allow_multiple = int(allow_multiple)
        msg2.pid = os.getpid()
        msg2.mod_id = self.module_id
        msg2.name = self.name

        self.send_message(msg2)
        self.send_message(msg)
        ack_msg = self._wait_for_acknowledgement()

        # save own module ID from ACK if asked to be assigned dynamic ID
        if self._module_id == 0:
            self._module_id = ack_msg.header.dest_mod_id

        # reset subscribed and paused types
        self._subscribed_types = set()
        self._paused_types = set()
        self._sub_all = False

        self.logger.log_name = self._name or f"Module {self._module_id}"

        return ack_msg

    @property
    def log_level(self) -> int:
        return self._logger.level

    @log_level.setter
    def log_level(self, level: int):
        self._logger.set_all_levels(level)

    @property
    def name(self) -> str:
        return self._name

    def connect(
        self,
        server_name: str = "localhost:7111",
        logger_status: bool = False,
        daemon_status: bool = False,
        allow_multiple: bool = False,
    ):
        """Connect to message manager server

        Args:
            server_name (optional): IP_addr:port_num string associated with message manager.
                Defaults to "localhost:7111".
            logger_status (optional): Flag to declare client as a logger module.
                Logger modules are automatically subscribed to all message types.
                Defaults to False.
            allow_multiple (optional): Flag to declare client can have multiple instances. Defaults to False.

        Raises:
            MessageManagerNotFound: Unable to connect to message manager
        """
        # Disconnect first
        if self.connected:
            self.disconnect()

        # Setup the underlying socket connection
        self._socket_connect(server_name)

        ack = self._connect_helper(logger_status, daemon_status, allow_multiple)

        # Send client pid
        self.send_module_ready()

    def disconnect(self):
        """Disconnect from message manager server"""
        try:
            if self._connected:
                self.send_signal(cd.MT_DISCONNECT)
                # Allow some time for signal to reach MM
                time.sleep(0.100)
        except:
            pass
        finally:
            if hasattr(self, "_sock"):
                self._sock.close()
            self._connected = False
            # reset subscribed and paused types
            self._subscribed_types = set()
            self._paused_types = set()
            self._sub_all = False

    @property
    def server(self) -> Tuple[str, int]:
        """Message manager server address as a (IP_addr, port_num) tuple"""
        return self._server

    @property
    def ip_addr(self) -> str:
        """Message manager IP address string"""
        return self._server[0]

    @property
    def port(self) -> int:
        """Message manager port number"""
        return self._server[1]

    @property
    def connected(self) -> bool:
        """Status of connection to message manager server"""
        return self._connected

    @property
    def msg_count(self) -> int:
        """Count of messages that have been sent"""
        return self._msg_count

    @property
    def module_id(self) -> int:
        """Numeric module ID of client"""
        return self._module_id

    @property
    def sock(self) -> socket.socket:
        """Underlying socket connection with MessageManager"""
        return self._sock

    @property
    def header_cls(self) -> Type[MessageHeader]:
        """Class defining the RTMA message header"""
        return self._header_cls

    @property
    def subscribed_types(self) -> Set[int]:
        """List of subscribed message types"""
        return set(self._subscribed_types)

    @property
    def paused_subscribed_types(self) -> Set[int]:
        """Subscriptions on pause"""
        return set(self._paused_types)

    @requires_connection
    def send_module_ready(self):
        """Send a signal to message manager that client is ready

        This method also sends the client's process ID to message manager.
        """
        msg = cd.MDF_MODULE_READY()
        msg.pid = os.getpid()
        self.send_message(msg)

    def _subscription_control(self, msg_list: Iterable[int], ctrl_msg: str):
        all_msg = ALL_MESSAGE_TYPES in msg_list

        if not all_msg:
            msg_set = set(msg_list)
        else:
            msg_set = set([ALL_MESSAGE_TYPES])

        # Note: Ignore any sub control for individual msg_types
        # when subscribed to ALL_MESSAGE_TYPES
        if self._sub_all and not all_msg:
            raise InvalidSubscription(
                "Cannot modify subscriptions to individual MTs while subscribed to ALL_MESSAGE_TYPES"
            )

        msg: MessageData
        if ctrl_msg == "Subscribe":
            msg = cd.MDF_SUBSCRIBE()
            if all_msg:
                self._subscribed_types = msg_set
                self._paused_types.clear()
                self._sub_all = True
            else:
                self._subscribed_types |= msg_set
                self._paused_types -= msg_set
        elif ctrl_msg == "Unsubscribe":
            msg = cd.MDF_UNSUBSCRIBE()
            if all_msg:
                self._subscribed_types.clear()
                self._paused_types.clear()
                self._sub_all = False
            else:
                self._subscribed_types -= msg_set
                self._paused_types -= msg_set
        elif ctrl_msg == "PauseSubscription":
            msg = cd.MDF_PAUSE_SUBSCRIPTION()
            if all_msg:
                self._subscribed_types.clear()
                self._paused_types.clear()
                self._sub_all = False
            else:
                self._subscribed_types -= msg_set
                self._paused_types |= msg_set
        elif ctrl_msg == "ResumeSubscription":
            msg = cd.MDF_RESUME_SUBSCRIPTION()
            if all_msg:
                self._subscribed_types = msg_set
                self._paused_types.clear()
                self._sub_all = True
            else:
                self._subscribed_types |= msg_set
                self._paused_types -= msg_set
        else:
            raise TypeError("Unknown control message type.")

        for msg_type in msg_set:
            msg.msg_type = msg_type
            self.send_message(msg)

    @requires_connection
    def subscribe(self, msg_list: Iterable[int]):
        """Subscribe to message types

        Calling this method multiple times will add to, and not replace,
        the list of subscribed messages.

        Args:
            msg_list (Iterable[int]): A list of numeric message IDs to subscribe to
        """
        self._subscription_control(msg_list, "Subscribe")

    @requires_connection
    def unsubscribe(self, msg_list: Iterable[int]):
        """Unsubscribe from message types

        Args:
            msg_list (Iterable[int]): A list of numeric message IDs to unsubscribe to
        """
        self._subscription_control(msg_list, "Unsubscribe")

    @requires_connection
    def pause_subscription(self, msg_list: Iterable[int]):
        """Pause subscription to message types

        Args:
            msg_list (Iterable[int]): A list of numeric message IDs to temporarily unsubscribe to
        """
        self._subscription_control(msg_list, "PauseSubscription")

    @requires_connection
    def resume_subscription(self, msg_list: Iterable[int]):
        """Resume subscription to message types

        Args:
            msg_list (Iterable[int]): A list of paused message IDs to resubscribe to
        """
        self._subscription_control(msg_list, "ResumeSubscription")

    @requires_connection
    def unsubscribe_from_all(self):
        """Unsubscribe from all subscribed types"""
        self.unsubscribe(self.subscribed_types)

    @requires_connection
    def pause_all_subscriptions(self):
        """Pause all subscribed types"""
        self.pause_subscription(self.subscribed_types)

    @requires_connection
    def resume_all_subscriptions(self):
        """Resume all paused subscriptions"""
        self.resume_subscription(self.paused_subscribed_types)

    @contextmanager
    def subscription_context(self, msg_list: Iterable[int]):
        """Context manager to subscribe to a list of message types

        Message types will automatically unsubscribe after exiting context.

        Args:
            msg_list (Iterable[int]): A list of numeric message IDs to subscribe to
        """
        msg_list = list(msg_list)  # cast arbitrary iterable to list
        for mt in msg_list:
            if mt in self.subscribed_types:
                warn(
                    f"Message ID {mt} is already subscribed, ignored from subscription_context"
                )
                msg_list.remove(mt)

        self.subscribe(msg_list)
        yield
        self.unsubscribe(msg_list)

    @contextmanager
    def paused_subscription_context(self, msg_list: Iterable[int]):
        """Context manager to pause subscriptions to a list of message types

        Message types will automatically resume subscriptions after exiting context.

        Args:
            msg_list (Iterable[int]): A list of numeric message IDs to temporarily unsubscribe to
        """

        msg_list = list(msg_list)  # cast arbitrary iterable to list
        for mt in msg_list:
            if mt not in self.subscribed_types:
                warn(
                    f"Message ID {mt} is not subscribed, ignored from paused_subscription_context"
                )
                msg_list.remove(mt)

        self.pause_subscription(msg_list)
        yield
        self.resume_subscription(msg_list)

    @requires_connection
    def send_signal(
        self,
        signal_type: int,
        dest_mod_id: int = 0,
        dest_host_id: int = 0,
        timeout: float = -1,
    ):
        """Send a signal

        A signal is a message type without an associated data payload.
        Only a unique message type ID is required to send a signal.
        To send a message with data, see :py:func:`send_message`.

        Args:
            signal_type: Numeric message type ID of signal
            dest_mod_id (optional): Specific module ID to send to. Defaults to 0 (broadcast).
            dest_host_id (optional): Specific host ID to send to. Defaults to 0 (broadcast).
            timeout (optional): Timeout in seconds to wait for socket to be available for sending.
                Defaults to -1 (blocking).

        Raises:
            InvalidDestinationModule: Specified destination module is invalid
            InvalidDestinationHost: Specified destination host is invalid
        """
        # Verify that the module & host ids are valid
        if dest_mod_id < 0 or dest_mod_id > cd.MAX_MODULES:
            raise InvalidDestinationModule(f"Invalid dest_mod_id  of [{dest_mod_id}]")

        if dest_host_id < 0 or dest_host_id > cd.MAX_HOSTS:
            raise InvalidDestinationHost(f"Invalid dest_host_id of [{dest_host_id}]")

        if timeout >= 0:
            readfds, writefds, exceptfds = select.select([], [self._sock], [], timeout)
        else:
            readfds, writefds, exceptfds = select.select(
                [], [self._sock], []
            )  # blocking

        if writefds:
            header = self._header_cls()
            header.msg_type = signal_type
            header.msg_count = self._msg_count
            header.send_time = time.perf_counter()
            header.recv_time = 0.0
            header.src_host_id = self._host_id
            header.src_mod_id = self._module_id
            header.dest_host_id = dest_host_id
            header.dest_mod_id = dest_mod_id
            header.num_data_bytes = 0
            header.remaining_bytes = 0
            header.reserved = 0

            self._sendall(header)  # type: ignore

            self._msg_count += 1

        else:
            # Socket was not ready to receive data. Drop the packet.
            print("x", end="")

    @requires_connection
    def send_message(
        self,
        msg_data: MessageData,
        dest_mod_id: int = 0,
        dest_host_id: int = 0,
        timeout: float = -1,
    ):
        """Send a message

        A message is a packet that contains a defined data payload.
        To send a message without associated data, see :py:func:`send_signal`.

        Args:
            msg_data: Object containing the message to send
            dest_mod_id (optional): Specific module ID to send to. Defaults to 0 (broadcast).
            dest_host_id (optional): Specific host ID to send to. Defaults to 0 (broadcast).
            timeout (optional): Timeout in seconds to wait for socket to be available for sending.
                Defaults to -1 (blocking).

        Raises:
            InvalidDestinationModule: Specified destination module is invalid
            InvalidDestinationHost: Specified destination host is invalid
        """
        # Verify that the module & host ids are valid
        if dest_mod_id < 0 or dest_mod_id > cd.MAX_MODULES:
            raise InvalidDestinationModule(f"Invalid dest_mod_id of [{dest_mod_id}]")

        if dest_host_id < 0 or dest_host_id > cd.MAX_HOSTS:
            raise InvalidDestinationHost(f"Invalid dest_host_id of [{dest_host_id}]")

        if timeout >= 0:
            readfds, writefds, exceptfds = select.select([], [self._sock], [], timeout)
        else:
            readfds, writefds, exceptfds = select.select(
                [], [self._sock], []
            )  # blocking

        if writefds:
            with disable_message_validation():
                header = self._header_cls()
                header.msg_type = msg_data.type_id
                header.msg_count = self._msg_count
                header.send_time = time.perf_counter()
                header.recv_time = 0.0
                header.src_host_id = self._host_id
                header.src_mod_id = self._module_id
                header.dest_host_id = dest_host_id
                header.dest_mod_id = dest_mod_id
                header.num_data_bytes = ctypes.sizeof(msg_data)
                header.remaining_bytes = 0
                header.reserved = 0

                try:
                    header.version = msg_data.type_hash
                except AttributeError as e:
                    if not hasattr(msg_data, "type_hash"):
                        warn(
                            "Message class is missing type_hash. V1 message defs are deprecated.",
                            FutureWarning,
                        )
                    else:
                        raise e

            self._sendall(header)  # type: ignore
            if header.num_data_bytes > 0:
                self._sendall(msg_data)  # type: ignore

            self._msg_count += 1

        else:
            # Socket was not ready to receive data. Drop the packet.
            print("x", end="")

    @requires_connection
    def forward_message(
        self,
        msg_hdr: MessageHeader,
        msg_data: Optional[MessageData] = None,
        timeout: float = -1,
    ):
        """Forward a message

        A message is a packet that contains a defined data payload.
        To send a message without associated data, see :py:func:`send_signal`.

        Args:
            msg_hdr: Object containing RTMA header to send
            msg_data: Object containing the message to send
            timeout (optional): Timeout in seconds to wait for socket to be available for sending.
                Defaults to -1 (blocking).

        """
        # Assume that msg_type, num_data_bytes, data - have been filled in
        if msg_data is not None:
            msg_hdr.num_data_bytes = ctypes.sizeof(msg_data)

        if timeout >= 0:
            readfds, writefds, exceptfds = select.select([], [self._sock], [], timeout)
        else:
            readfds, writefds, exceptfds = select.select(
                [], [self._sock], []
            )  # blocking

        if writefds:
            self._sendall(msg_hdr)  # type: ignore
            if msg_data is not None:
                self._sendall(msg_data)  # type: ignore

            self._msg_count += 1

        else:
            # Socket was not ready to write data. Drop the packet.
            print("x", end="")

    def _sendall(self, buffer: bytearray):
        try:
            self._sock.sendall(buffer)
        except ConnectionError as e:
            self._connected = False
            raise ConnectionLost from e

    @requires_connection
    def read_message(
        self, timeout: Union[int, float, None] = -1, ack=False, sync_check=False
    ) -> Optional[Message]:
        """Read a message

        Args:
            timeout (optional): Timeout to wait for a message to be available for reading.
                Defaults to -1 (blocking).
            ack (optional): Primarily for internal use. When True, will not discard ACK messages. Defaults to False.
            sync_check (optional): Validate message definition matches header version. Defaults to False.

        Raises:
            ConnectionLost: Connection error to message manager server

        Returns:
            Message object. If no message is read before timeout, returns None.
        """
        t0 = time.perf_counter()
        M = self._read_message(timeout, ack, sync_check)

        if self._sub_all:
            return M

        # filter out unsubscribed messages that may still be in queue
        while M and (M.header.msg_type not in self.subscribed_types):
            if (
                ack and M.header.msg_type == cd.MT_ACKNOWLEDGE
            ):  # ack input allows ACK to be read and not discarded
                break
            if timeout is None:  # blocking
                t_rem = None  # continue blocking until subscribed message recv'd
            elif timeout == 0:  # no timeout
                M = None  # discard unsub'd msg and break
                break
            else:  # subtract time elapsed from timeout and read again
                t_rem = max(timeout - (time.perf_counter() - t0), 0)
            M = self._read_message(timeout, ack, sync_check)

        return M

    @requires_connection
    def _read_message(
        self, timeout: Union[int, float, None] = -1, ack=False, sync_check=False
    ) -> Optional[Message]:
        """Read message without filtering for subscribed messages (helper called by read_message)"""

        if timeout is None:
            # Skip select call
            pass
        elif timeout >= 0:
            # Wait timeout amount
            readfds, writefds, exceptfds = select.select([self._sock], [], [], timeout)
            if len(readfds) == 0:
                return None
        else:
            # Blocking
            readfds, writefds, exceptfds = select.select([self._sock], [], [])
            if len(readfds) == 0:
                return None

        # Read RTMA Header Section
        header = self._header_cls()
        try:
            nbytes = self._sock.recv_into(header, header.size, socket.MSG_WAITALL)
            """
            Note:
            MSG_WAITALL Flag:
            The receive request will complete only when one of the following events occurs:
            The buffer supplied by the caller is completely full.
            The connection has been closed.
            The request has been canceled or an error occurred.
            """

            if nbytes != header.size:
                self._connected = False
                raise ConnectionLost

            header.recv_time = time.perf_counter()
        except ConnectionError:
            raise ConnectionLost

        # Read Data Section

        try:
            data = get_msg_cls(header.msg_type)()
        except UnknownMessageType as e:
            mt = header.msg_type
            raw = self._sock.recv(header.num_data_bytes, socket.MSG_WAITALL)
            raise UnknownMessageType(
                f"No message definition found for MT={mt}", header, raw
            )

        type_size = data.type_size
        if type_size == -1:  # not defined for v1 message defs
            type_size = data.size

        if type_size != header.num_data_bytes:
            _ = self._sock.recv(header.num_data_bytes, socket.MSG_WAITALL)
            raise InvalidMessageDefinition(
                f"Received message header indicating a message data size ({header.num_data_bytes}) that does not match the expected size ({type_size}) of message type {data.type_name}. Message definitions may be out of sync across systems."
            )

        # Note: Ignore the sync check if header.version is not filled in
        # This can removed once all clients support this field.
        if sync_check and header.version != 0 and header.version != data.type_hash:
            _ = self._sock.recv(header.num_data_bytes, socket.MSG_WAITALL)
            raise InvalidMessageDefinition(
                f"Received message header indicating a message version that does not match the expected version of message type {data.type_name}. Message definitions may be out of sync across systems."
            )

        if header.num_data_bytes:
            try:
                nbytes = self._sock.recv_into(data, type_size, socket.MSG_WAITALL)

                if nbytes != type_size:
                    self._connected = False
                    raise ConnectionLost
            except ConnectionError:
                raise ConnectionLost

        return Message(header, data)

    def _wait_for_acknowledgement(self, timeout: float = 3) -> Message:
        """Wait for acknowledgement from message manager module

        Used internally when acknowledgement replies from message manager are expected

        Args:
            timeout (optional): Timeout in seconds to wait for ack message. Defaults to 3.

        Raises:
            AcknowledgementTimeout: Ack not received from message manager

        Returns:
            Ack message
        """
        ret = 0

        # Wait Forever
        if timeout == -1:
            while True:
                msg = self.read_message(ack=True)
                if msg is not None:
                    if msg.header.msg_type == cd.MT_ACKNOWLEDGE:
                        break
            return msg
        else:
            # Wait up to timeout seconds
            time_remaining = timeout
            start_time = time.perf_counter()
            while time_remaining > 0:
                msg = self.read_message(timeout=time_remaining, ack=True)
                if msg is not None:
                    if msg.header.msg_type == cd.MT_ACKNOWLEDGE:
                        return msg

                time_now = time.perf_counter()
                time_waited = time_now - start_time
                time_remaining = timeout - time_waited

            raise AcknowledgementTimeout(
                "Failed to receive Acknowlegement from MessageManager"
            )

    def discard_messages(self, timeout: float = 1) -> bool:
        """Read and discard messages in socket buffer up to timeout

        Args:
            timeout (optional): Maximum time in seconds to loop through message buffer.
                Defaults to 1.

        Returns:
            True if all messages have been read, False if messages remain in buffer
        """

        """Read and discard messages in socket buffer up to timeout.
        Returns: True if all messages available have been read.
        """
        start_time = time.perf_counter()

        def read_and_time():
            msg = self.read_message(0)
            time_now = time.perf_counter()
            time_waited = time_now - start_time
            time_remaining = timeout - time_waited
            return msg, time_remaining

        msg, time_remaining = read_and_time()
        while msg is not None and time_remaining > 0:
            msg, time_remaining = read_and_time()
        return not msg

    def reset_subscriptions(self, subscribe_list: Optional[Iterable[int]] = None):
        """Unsubscribe from all messages and empty the message buffer.
        If optional 'subscribe_list' argument given, then will subscribe to new messages after reset.

        Args:
            subscribe_list (Optional[Iterable[int]], optional): List of message IDs to subscribe to. Defaults to None.
        """
        self.unsubscribe_from_all()
        self.discard_messages()
        if subscribe_list:
            self.subscribe(subscribe_list)

    def __str__(self) -> str:
        # TODO: Make this better.
        return f"Client(module_id={self.module_id}, server={self.server}, connected={self.connected})"


@contextmanager
def client_context(
    module_id: int = 0,
    server_name: str = "localhost:7111",
    msg_list: Optional[Iterable[int]] = None,
    host_id: int = 0,
    timecode: bool = False,
    logger_status: bool = False,
    allow_multiple: bool = False,
    name: str = "",
):
    """Context manager function to simplify initializing a pyrtma Client

    Context manager will yield a Client object after connecting to message manager,
    optionally subscribing to msg_list, and after calling send_module_ready().
    Client will disconnect when exiting context.

    Args:
        module_id (optional): Static module ID, which must be unique.
            Defaults to 0, which generates a dynamic module ID.
        server_name (optional): IP_addr:port_num string associated with message manager.
                Defaults to "localhost:7111".
        msg_list (optional): A list of numeric message IDs to subscribe to
        host_id (optional): Host ID. Defaults to 0.
        timecode (optional): Add additional timecode fields to message
            header, used by some projects at RNEL. Defaults to False.
        logger_status (optional): Flag to declare client as a logger module.
            Logger modules are automatically subscribed to all message types.
            Defaults to False.
        allow_multiple (optional): Flag to declare client can have multiple instances. Defaults to False.
        name (optional): Name of module

    Yields:
        Client: initialized pyrtma Client object
    """
    c = Client(module_id, host_id, timecode, name=name)
    c.connect(server_name, logger_status, allow_multiple)
    if msg_list:
        c.subscribe(msg_list)
    c.send_module_ready()
    yield c
    c.disconnect()



import pyrtma
from typing import Type, Optional, ClassVar, IO, Any, Union, List, Dict
from abc import ABC
from .exceptions import InvalidFormatter, DataFormatterKeyError


class DataFormatter(ABC):
    name: ClassVar[str] = "base"
    mode: ClassVar[str] = "wt"
    ext: ClassVar[str] = ".txt"

    def __init__(self, fd: IO[Any], **kwargs):
        self.fd = fd
        if hdr := self.format_header():
            self.fd.write(hdr)

    def format_header(self) -> Optional[Union[bytes, str]]:
        return None

    def format_footer(self) -> Optional[Union[bytes, str]]:
        return None

    def format_message(self, msg: pyrtma.Message) -> Optional[Union[str, bytes]]:
        return None

    def write(self, wbuf: List[pyrtma.Message]):
        self.fd.writelines(self.format_message(msg) for msg in wbuf)

    def finalize(self, wbuf: List[pyrtma.Message]):
        # Write the last bit of data
        self.write(wbuf)

        # Write an optional footer
        if footer := self.format_footer():
            self.fd.write(footer)


_formatter_map: Dict[str, Type[DataFormatter]] = {}


def add_formatter(fmt_cls: Type[DataFormatter]):
    assert issubclass(fmt_cls, DataFormatter)
    if _formatter_map.get(fmt_cls.name) is not None:
        raise DataFormatterKeyError(
            f"A DataFormatter already exists with the name: {fmt_cls.name}"
        )

    _formatter_map[fmt_cls.name] = fmt_cls


def get_formatter(name: str) -> Type[DataFormatter]:
    fmt_cls = _formatter_map.get(name)

    if fmt_cls is not None:
        return fmt_cls
    else:
        raise InvalidFormatter(f"No DataFormatter class named {name}")



import pathlib
import random
from ruamel.yaml import YAML
from typing import Dict, Any, Union

# Field type name to ctypes
type_map = [
    "char",
    "unsigned char",
    "float",
    "double",
    "uint8",
    "uint16",
    "uint32",
    "uint64",
    "int8",
    "int16",
    "int32",
    "int64",
]


class TestDefGenerator:
    """Class to generate random message defs for testing"""

    def __init__(self):
        self._id = 1000
        self.constants = {}
        self.struct_defs = {}
        self.message_defs = {}

    @property
    def id(self) -> int:
        self._id += 1
        return self._id

    def random_nfields(self) -> int:
        """Randomly pick number of fields between 0-5

        Returns:
            int: Random int 0-5
        """
        return random.randint(0, 5)

    def random_flen(self) -> Union[str, int]:
        """Random field length

        Returns:
            Union[str, int]: Random field length, either 0 or a random expression
        """
        if random.random() > 0.5:
            return 0
        return self.random_expression()

    def random_value(self) -> int:
        """Random int value 1-4

        Returns:
            int: Random int 1-4
        """
        return random.randint(1, 4)

    def random_op(self) -> str:
        """Random operation

        Returns:
            str: Randomly add, multiply or no operation
        """
        return random.choice(["", "+", "*"])

    def random_constant(self) -> Union[str, int]:
        """Randomly pick a constant

        Returns:
            Union[str, int]: Constant expression or random int value
        """
        if len(self.constants):
            return random.choice(list(self.constants.keys()))
        else:
            return self.random_value()

    def random_expression(self) -> Union[str, int]:
        """Random expression

        Returns:
            Union[str, int]: Random expression (may include math operation) or int value
        """
        if random.random() > 0.66:
            c = self.random_constant()
        else:
            c = self.random_value()

        if type(c) is str:
            op = self.random_op()
            if op:
                c2 = self.random_value()
                return f"{c} {op} {c2}"

        return c

    def random_ftype(self, nested: bool = False) -> str:
        """Random field type name

        Args:
            nested (bool, optional): Flag for nested type. Defaults to False.

        Returns:
            str: Field type name
        """
        if not nested:
            return random.choice(type_map)

        v = random.random()
        if v < 0.75:
            return random.choice(type_map)

        if random.random() > 0.5:
            if len(self.struct_defs):
                s = list(k for (k, v) in self.struct_defs.items() if len(v["fields"]))
            else:
                s = []
            return random.choice(s or type_map)
        else:
            if len(self.message_defs):
                m = list(k for (k, v) in self.message_defs.items() if v["fields"])
            else:
                m = []
            return random.choice(m or type_map)

    def generate_constants(self, n: int):
        """Generate random constants

        Args:
            n (int): Number of constants to generate
        """
        for i in range(1, n):
            name = f"C_{i:02d}"
            self.constants[name] = self.random_expression()

    def generate_structs(self, n: int):
        """Generate random structs

        Args:
            n (int): Number of structs to generate
        """
        for i in range(1, n):
            name = f"SRCT_{i:02d}"
            fields = {}
            nfields = self.random_nfields() or 1
            for k in range(1, nfields + 1):
                ftype = self.random_ftype()
                fname = f"f{k}"
                flen = self.random_flen()
                if flen == 0:
                    fields[fname] = f"{ftype}"
                else:
                    fields[fname] = f"{ftype}[{flen}]"

            self.struct_defs[name] = dict(fields=fields)

    def generate_message_defs(self, n: int):
        """Generate random message defs

        Args:
            n (int): Number of message defs to generate
        """
        for i in range(1, n):
            name = f"MSG_{i:03d}"
            fields = {}
            nfields = self.random_nfields()
            for k in range(1, nfields + 1):
                ftype = self.random_ftype(nested=True)
                fname = f"f{k}"
                flen = self.random_flen()
                if flen == 0:
                    fields[fname] = f"{ftype}"
                else:
                    fields[fname] = f"{ftype}[{flen}]"

            self.message_defs[name] = dict(id=self.id, fields=fields or None)

    def generate(self, out_filepath: pathlib.Path):
        """Generate yaml file with random test message defs

        Args:
            out_filepath (pathlib.Path): output filepath
        """
        self.generate_constants(24)
        self.generate_structs(10)
        self.generate_message_defs(256)

        d: Dict[str, Any] = dict(
            constants=self.constants,
            struct_defs=self.struct_defs,
            message_defs=self.message_defs,
        )

        class NullRepresenter:
            def __init__(self):
                self.count = 0

            def __call__(self, repr, data):
                return repr.represent_scalar("tag:yaml.org,2002:null", "null")

        yaml = YAML()
        yaml.default_flow_style = False
        my_represent_none = NullRepresenter()
        yaml.representer.add_representer(type(None), my_represent_none)
        yaml.dump(d, out_filepath)


if __name__ == "__main__":
    g = TestDefGenerator()
    d = g.generate(pathlib.Path("./test_defs.yaml"))



"""pyrtma.constants module

This module is deprecated and included for backwards compatilibity with pyrtma 1.2.3
Constants have been moved to pyrtma.core_defs
"""

from .core_defs import (
    MAX_MODULES,
    DYN_MOD_ID_START,
    MAX_HOSTS,
    MAX_MESSAGE_TYPES,
    MAX_RTMA_MSG_TYPE,
    MAX_RTMA_MODULE_ID,
    MAX_LOGGER_FILENAME_LENGTH,
    MAX_CONTIGUOUS_MESSAGE_DATA,
    ALL_MESSAGE_TYPES,
    MODULE_ID,
    HOST_ID,
    MSG_TYPE,
    MSG_COUNT,
    LOCAL_HOST,
    ALL_HOSTS,
    MID_MESSAGE_MANAGER,
    MID_QUICK_LOGGER,
    MT_EXIT,
    MT_KILL,
    MT_ACKNOWLEDGE,
    MT_FAIL_SUBSCRIBE,
    MT_FAILED_MESSAGE,
    MT_CONNECT,
    MT_DISCONNECT,
    MT_SUBSCRIBE,
    MT_UNSUBSCRIBE,
    MT_PAUSE_SUBSCRIPTION,
    MT_RESUME_SUBSCRIPTION,
    MT_SAVE_MESSAGE_LOG,
    MT_MESSAGE_LOG_SAVED,
    MT_FORCE_DISCONNECT,
    MT_MODULE_READY,
    MT_TIMING_MESSAGE,
)

from warnings import warn

warn(
    "pyrtma.constants module is deprecated and has been replaced by pyrtma.core_defs",
    FutureWarning,
)

HID_LOCAL_HOST = LOCAL_HOST
HID_ALL_HOSTS = ALL_HOSTS

# DEPRECATED MIDS
MID_COMMAND_MODULE = 1
MID_APPLICATION_MODULE = 2
MID_NETWORK_RELAY = 3
MID_STATUS_MODULE = 4
MID_QUICKLOGGER = MID_QUICK_LOGGER

# DEPRECATED MTs
MT_PAUSE_MESSAGE_LOGGING = 58  # not fully implemented
MT_RESUME_MESSAGE_LOGGING = 59  # not fully implemented
MT_RESET_MESSAGE_LOG = 60  # un-deprecate?
MT_DUMP_MESSAGE_LOG = 61  # un-deprecate?



import pathlib

from textwrap import dedent
from typing import Union
from pyrtma.parser import (
    Parser,
    ConstantExpr,
    ConstantString,
    TypeAlias,
    MT,
    MID,
    HID,
    MDF,
    SDF,
)
from pyrtma.__version__ import __version__

tab = "    "

# Native C types that are supported
type_map = {
    "char": "char",
    "unsigned char": "unsigned char",
    "byte": "uint8_t",
    "int": "int32_t",
    "signed int": "int32_t",
    "unsigned int": "uint32_t",
    "unsigned": "uint32_t",
    "short": "int16_t",
    "signed short": "int16_t",
    "unsigned short": "uint16_t",
    "long": "int32_t",
    "signed long": "int32_t",
    "unsigned long": "uint32_t",
    "long long": "int64_t",
    "signed long long": "int64_t",
    "unsigned long long": "uint64_t",
    "float": "float",
    "double": "double",
    "uint8": "uint8_t",
    "uint16": "uint16_t",
    "uint32": "uint32_t",
    "uint64": "uint64_t",
    "int8": "int8_t",
    "int16": "int16_t",
    "int32": "int32_t",
    "int64": "int64_t",
}


def pad(indent: int) -> str:
    return indent * "\t"


class CDefCompiler:
    def __init__(self, parser: Parser, filename: str, debug: bool = False):
        self.debug = debug
        self.parser = parser
        self.filename = filename

    def generate_constant(self, c: ConstantExpr):
        return f"#define {c.name:<48}{c.value}\n"

    def generate_string_constant(self, c: ConstantString):
        return f"#define {c.name:<48}{c.value}\n"

    def generate_msg_type_id(self, mt: MT) -> str:
        return f"#define MT_{mt.name:<48}{mt.value}\n"

    def generate_host_id(self, hid: HID) -> str:
        return f"#define HID_{hid.name:<48}{hid.value}\n"

    def generate_module_id(self, mid: MID) -> str:
        return f"#define MID_{mid.name:<48}{mid.value}\n"

    def generate_hash_id(self, mdf: Union[MDF, SDF]) -> str:
        return f"#define HASH_{mdf.name:<48} 0x{mdf.hash[:8]}\n"

    def generate_type_alias(self, td: TypeAlias) -> str:
        if td.type_name in type_map.keys():
            return f"typedef {type_map[td.type_name]} {td.name};\n"

        if td.type_name in self.parser.aliases.keys():
            return f"typedef {td.type_name} {td.name};\n"

        if td.type_name in self.parser.struct_defs.keys():
            return f"typedef {td.type_name} {td.name};\n"

        if td.type_name in self.parser.message_defs.keys():
            return f"typedef MDF_{td.type_name} MDF_{td.name};\n"

        raise RuntimeError(f"No type found for alias: {td.name}")

    def generate_struct(self, struct: Union[SDF, MDF]) -> str:
        if isinstance(struct, MDF):
            prefix = "MDF_"
        elif isinstance(struct, SDF):
            prefix = ""

        if len(struct.fields) == 0:
            return f"// {struct.name} -> Signal"

        f = ["typedef struct {"]
        tabs = f"{tab}" * 1
        for field in struct.fields:
            if field.type_name in type_map.keys():
                ftype = f"{type_map[field.type_name]}"
            elif field.type_name in self.parser.message_defs.keys():
                ftype = f"MDF_{field.type_name}"
            elif field.type_name in self.parser.struct_defs.keys():
                ftype = f"{field.type_name}"
            elif field.type_name in self.parser.aliases.keys():
                ftype = f"{field.type_name}"
            else:
                raise RuntimeError(f"Unknown field name {field.name} in {struct.name}")

            if field.length is None:
                f.append(f"{tabs}{ftype} {field.name};")
            else:
                f.append(f"{tabs}{ftype} {field.name}[{field.length}];")

        f.append(f"}} {prefix}{struct.name};\n")

        return "\n".join(f)

    def generate_version_comment(self) -> str:
        s = f"""\
            // This message def file was auto-generated by pyrtma.compile version {__version__}

            """

        return dedent(s)

    def generate_includes(self) -> str:
        filename = self.filename.replace(" ", "_").upper()
        s = f"""\
            #ifndef _{filename}_
            #define _{filename}_

            // Sized integer types
            #include <stdint.h>

            """

        return dedent(s)

    def generate_type_info(self) -> str:
        s = ""

        s += "typedef struct {\n"
        obj: Union[ConstantExpr, ConstantString, HID, MID, TypeAlias, SDF, MT, MDF]
        for obj in self.parser.constants.values():
            if obj.src.parent.stem == "core_defs":
                continue  # exclude core_defs.yaml, C clients will import rtma.h
            if type(obj.value) is float:
                ftype = "double"
            else:
                ftype = "int"

            s += f'#pragma push_macro("{obj.name}")\n'
            s += f"#undef {obj.name}\n"
            s += f"    {ftype} {obj.name};\n"
            s += f'#pragma pop_macro("{obj.name}")\n'
            s += "\n"

        for obj in self.parser.string_constants.values():
            if obj.src.parent.stem == "core_defs":
                continue  # exclude core_defs.yaml, C clients will import rtma.h
            s += f'#pragma push_macro("{obj.name}")\n'
            s += f"#undef {obj.name}\n"
            s += f"    const char* {obj.name};\n"
            s += f'#pragma pop_macro("{obj.name}")\n'
            s += "\n"

        s += "} ConstantInfo;\n"
        s += "\n" * 2

        s += "typedef struct {\n"
        for obj in self.parser.host_ids.values():
            if obj.src.parent.stem == "core_defs":
                continue  # exclude core_defs.yaml, C clients will import rtma.h
            s += f"    HOST_ID {obj.name};\n"
        s += "} HostIdInfo;\n"

        s += "\n" * 2

        s += "typedef struct {\n"
        for obj in self.parser.module_ids.values():
            if obj.src.parent.stem == "core_defs":
                continue  # exclude core_defs.yaml, C clients will import rtma.h
            s += f"    MODULE_ID {obj.name};\n"
        s += "} ModuleIdInfo;\n"

        s += "\n" * 2

        s += "typedef struct {\n"
        for obj in self.parser.message_ids.values():
            if obj.src.parent.stem == "core_defs":
                continue  # exclude core_defs.yaml, C clients will import rtma.h
            s += f"    MSG_TYPE {obj.name};\n"
        s += "} MsgIdInfo;\n"

        s += "\n" * 2

        s += "typedef struct {\n"
        for obj in self.parser.message_defs.values():
            if obj.src.parent.stem == "core_defs":
                continue  # exclude core_defs.yaml, C clients will import rtma.h
            s += f"    int {obj.name};\n"
        s += "} MsgHashInfo;\n"

        s += "\n" * 2

        s += "typedef struct {\n"
        s += "    ConstantInfo constants;\n"
        s += "    HostIdInfo HID;\n"
        s += "    ModuleIdInfo MID;\n"
        s += "    MsgIdInfo MT;\n"
        s += "    MsgHashInfo HASH;\n"
        s += "} RTMAInfo;\n"

        s += "\n" * 2

        return dedent(s)

    def generate_rtma_info_getter(self) -> str:
        s = "const RTMAInfo get_rtma_info() {\n"
        s += "    static RTMAInfo RTMA;\n"
        obj: Union[ConstantExpr, ConstantString, HID, MID, TypeAlias, SDF, MT, MDF]
        for obj in self.parser.constants.values():
            if obj.src.parent.stem == "core_defs":
                continue  # exclude core_defs.yaml, C clients will import rtma.h
            s += f'#pragma push_macro("{obj.name}")\n'
            s += f"#undef {obj.name}\n"
            s += f"{tab}RTMA.constants.{obj.name} = {obj.value};\n"
            s += f'#pragma pop_macro("{obj.name}")\n'
            s += "\n"

        for obj in self.parser.string_constants.values():
            if obj.src.parent.stem == "core_defs":
                continue  # exclude core_defs.yaml, C clients will import rtma.h
            s += f'#pragma push_macro("{obj.name}")\n'
            s += f"#undef {obj.name}\n"
            s += f"{tab}RTMA.constants.{obj.name} = {obj.value};\n"
            s += f'#pragma pop_macro("{obj.name}")\n'
            s += "\n"

        for obj in self.parser.host_ids.values():
            if obj.src.parent.stem == "core_defs":
                continue  # exclude core_defs.yaml, C clients will import rtma.h
            s += f"{tab}RTMA.HID.{obj.name} = {obj.value};\n"

        s += "\n" * 2

        for obj in self.parser.module_ids.values():
            if obj.src.parent.stem == "core_defs":
                continue  # exclude core_defs.yaml, C clients will import rtma.h
            s += f"{tab}RTMA.MID.{obj.name} = {obj.value};\n"

        s += "\n" * 2

        for obj in self.parser.message_ids.values():
            if obj.src.parent.stem == "core_defs":
                continue  # exclude core_defs.yaml, C clients will import rtma.h
            s += f"{tab}RTMA.MT.{obj.name} = {obj.value};\n"

        s += "\n" * 2

        for obj in self.parser.message_defs.values():
            if obj.src.parent.stem == "core_defs":
                continue  # exclude core_defs.yaml, C clients will import rtma.h
            s += f"{tab}RTMA.HASH.{obj.name} = 0x{obj.hash[:8]};\n"

        s += "\n"
        s += "    return RTMA;\n"
        s += "};\n"

        return s

    def generate_close_guard(self) -> str:
        filename = self.filename.replace(" ", "_").upper()
        return f"#endif // _{filename}_"

    def generate(self, out_filepath: pathlib.Path):
        with open(out_filepath, mode="w") as f:
            # Version comment
            f.write(self.generate_version_comment())

            # Includes
            f.write(self.generate_includes())
            # Note: above includes RTMA_type.h, and user code will typically include RTMA.h
            # For this reason, we will exclude compiling anything from core_defs.yaml, which is redundant with RTMA_types.h

            # Version
            f.write(f'#define COMPILED_PYRTMA_VERSION "{__version__}"\n\n')

            # Constants
            f.write("// Constants\n")
            obj: Union[ConstantExpr, ConstantString, HID, MID, TypeAlias, SDF, MT, MDF]
            for obj in self.parser.constants.values():
                # exclude core_defs.yaml, C clients will import rtma.h
                if obj.src.parent.stem != "core_defs":
                    f.write(self.generate_constant(obj))
            f.write("\n")

            # String Constants
            f.write("// String Constants\n")
            for obj in self.parser.string_constants.values():
                # exclude core_defs.yaml, C clients will import rtma.h
                if obj.src.parent.stem != "core_defs":
                    f.write(self.generate_string_constant(obj))
            f.write("\n")

            # Type Aliases
            f.write("// Type Aliases\n")
            for obj in self.parser.aliases.values():
                # exclude core_defs.yaml, C clients will import rtma.h
                if obj.src.parent.stem != "core_defs":
                    f.write(self.generate_type_alias(obj))
            f.write("\n")

            # Host IDs
            f.write("// Host IDs\n")
            for obj in self.parser.host_ids.values():
                # exclude core_defs.yaml, C clients will import rtma.h
                if obj.src.parent.stem != "core_defs":
                    f.write(self.generate_host_id(obj))
            f.write("\n")

            # Module IDs
            f.write("// Module IDs\n")
            for obj in self.parser.module_ids.values():
                # exclude core_defs.yaml, C clients will import rtma.h
                if obj.src.parent.stem != "core_defs":
                    f.write(self.generate_module_id(obj))
            f.write("\n")

            # Message Type IDs
            f.write("// Message Type IDs\n")
            for obj in self.parser.message_ids.values():
                # exclude core_defs.yaml, C clients will import rtma.h
                if obj.src.parent.stem != "core_defs":
                    f.write(self.generate_msg_type_id(obj))
            f.write("\n")

            # Structure Types
            f.write("// Struct Definitions\n")
            for obj in self.parser.struct_defs.values():
                # exclude core_defs.yaml, C clients will import rtma.h
                if obj.src.parent.stem != "core_defs":
                    f.write(self.generate_struct(obj))
                    f.write("\n\n")

            # Message Definitions
            f.write("// Message Definitions\n")
            for obj in self.parser.message_defs.values():
                # exclude core_defs.yaml, C clients will import rtma.h
                if obj.src.parent.stem != "core_defs":
                    f.write(self.generate_struct(obj))
                    f.write("\n\n")

            # Message Definition Hashes
            f.write("// Message Definition Hashes\n")
            for obj in self.parser.message_defs.values():
                # exclude core_defs.yaml, C clients will import rtma.h
                if obj.src.parent.stem != "core_defs":
                    f.write(self.generate_hash_id(obj))
            f.write("\n")

            # Message Definition Type Info (Future Use)
            # f.write("// Type Information\n")
            # f.write(self.generate_type_info())
            # f.write("\n")

            # TODO below commented out by JW because it causes link errors when header file is included multiple times (in multiple object files)
            # f.write(self.generate_rtma_info_getter())
            # f.write("\n")

            # Close header guard
            f.write(self.generate_close_guard())



import pathlib
import io
import logging
import math

from ..message import Message
from ..core_defs import ALL_MESSAGE_TYPES
from typing import Type, Optional, IO, Any, List

from .metadata import LoggingMetadata
from .data_formatter import DataFormatter
from .exceptions import DataSetExistsError


class DataSet:
    MIN_INTERVAL = 30
    MAX_INTERVAL = 600
    CONTINUOUS = math.inf

    def __init__(
        self,
        collection_name: str,
        name: str,
        sub_dir_fmt: str,
        file_name_fmt: str,
        formatter_cls: Type[DataFormatter],
        subdivide_interval: int,
        msg_types: List[int],
        metadata: LoggingMetadata,
    ):
        self.logger = logging.getLogger("data_logger").getChild(
            f"{collection_name}.{name}"
        )
        self.logger.name = f"{collection_name}.{name}"
        self.collection_name = collection_name
        self.name = name
        self.sub_dir_fmt = sub_dir_fmt
        self.file_name_fmt = file_name_fmt
        self.formatter_cls = formatter_cls

        if subdivide_interval <= 0:
            self.subdivide_interval = DataSet.CONTINUOUS
        elif subdivide_interval < DataSet.MIN_INTERVAL:
            self.subdivide_interval = DataSet.MIN_INTERVAL
        elif subdivide_interval > DataSet.MAX_INTERVAL:
            self.subdivide_interval = DataSet.MAX_INTERVAL
        else:
            self.subdivide_interval = subdivide_interval

        # Placeholder for type checking purposes
        if "b" in self.formatter_cls.mode:
            self.formatter = formatter_cls(io.BytesIO())
        else:
            self.formatter = formatter_cls(io.StringIO())

        self.rbuf: List[Message] = []
        self.wbuf: List[Message] = []

        self.fd: Optional[IO[Any]] = None
        self.file_path = pathlib.Path()
        self.metadata = metadata

        self.msg_types = [m for m in msg_types if m > 0]
        if ALL_MESSAGE_TYPES in msg_types:
            self.all_sub = True
        else:
            self.all_sub = False

        self.subdivide_flag = False
        self.next_subdivide = math.inf
        self.sub_index = 0
        self.collection_stopped = False
        self.base_file_name: str = ""

    def start(self, base_path: pathlib.Path):
        # Reset some tracking vars
        self.sub_index = 0
        self.collection_stopped = False
        self.subdivide_flag = False

        # Expand data set save directory
        save_dir = self.metadata.expand_format(self.sub_dir_fmt)

        save_path = base_path.joinpath(save_dir)
        save_path.mkdir(parents=True, exist_ok=True)

        # Expand data set file name
        filename = self.metadata.expand_format(
            self.file_name_fmt + self.formatter_cls.ext
        )

        self.file_path = save_path / filename

        if self.file_path.exists():
            raise DataSetExistsError(f"{self.file_path} already exists.")

        self.logger.info(f"Saving Data Set:{self.name} to {self.file_path}")

        self.fd = open(self.file_path, self.formatter_cls.mode)
        self.formatter = self.formatter_cls(self.fd)
        self.next_subdivide = self.subdivide_interval

    def subdivide(self):
        self.sub_index += 1
        self.formatter.finalize(self.wbuf)

        if self.fd:
            self.fd.close()

        base_name = self.metadata.expand_format(self.file_name_fmt)
        new_filename = f"{base_name}_{self.sub_index:04d}{self.formatter_cls.ext}"
        self.file_path = self.file_path.parent / new_filename

        if self.file_path.exists():
            raise DataSetExistsError(f"{self.file_path} already exists.")

        self.logger.info(f"Sub-dividing data set: {self.file_path}")

        self.fd = open(self.file_path, self.formatter_cls.mode)
        self.formatter = self.formatter_cls(self.fd)

    def stop(self):
        self.stage_for_write()
        self.formatter.finalize(self.wbuf)

    def close(self):
        if self.fd is not None:
            self.fd.close()

    def stage_for_write(self):
        """Set the write buffer data for the data collection write thread"""
        self.wbuf = self.rbuf
        self.rbuf = []

    def write(self):
        self.formatter.write(self.wbuf)
        self.wbuf.clear()

        if not self.collection_stopped and self.subdivide_flag:
            self.subdivide_flag = False
            self.subdivide()



import pathlib

from pyrtma.parser import (
    Parser,
    ConstantExpr,
    ConstantString,
    TypeAlias,
    MT,
    MID,
    HID,
    MDF,
    SDF,
)
from pyrtma.__version__ import __version__

from typing import Any, Union
from textwrap import dedent

# Field type name to matlab types
type_map = {
    "char": "int8",
    "unsigned char": "uint8",
    "byte": "uint8",
    "int": "int32",
    "signed int": "int32",
    "unsigned int": "uint32",
    "unsigned": "uint32",
    "short": "int16",
    "signed short": "int16",
    "unsigned short": "uint16",
    "long": "int32",
    "signed long": "int32",
    "unsigned long": "uint32",
    "long long": "int64",
    "signed long long": "int64",
    "unsigned long long": "uint64",
    "float": "single",
    "double": "double",
    "uint8": "uint8",
    "uint16": "uint16",
    "uint32": "uint32",
    "uint64": "uint64",
    "int8": "int8",
    "int16": "int16",
    "int32": "int32",
    "int64": "int64",
}


class MatlabDefCompiler:
    def __init__(self, parser: Parser, debug: bool = False):
        self.debug = debug
        self.parser = parser
        self.struct_name = "RTMA"

    def generate_fcn_header(self, fcn_name: str) -> str:
        return f"function {self.struct_name} = {fcn_name}()\n\n"

    def initialize_struct(self) -> str:
        s = f"""\
            {self.struct_name} = struct();
            {self.struct_name}.COMPILED_PYRTMA_VERSION = '{__version__}';
            {self.struct_name}.HID = [];
            {self.struct_name}.MID = [];
            {self.struct_name}.MT = [];
            {self.struct_name}.MDF = [];
            {self.struct_name}.MESSAGE_HEADER = [];
            {self.struct_name}.MTN_by_MT = [];
            {self.struct_name}.MDF_by_MT = [];
            {self.struct_name}.mex_opcode = [];
            {self.struct_name}.defines = [];
            {self.struct_name}.typedefs = [];
            {self.struct_name}.hash = [];
        """

        return dedent(s)

    @staticmethod
    def generate_fcn_close() -> str:
        return "end\n"

    @staticmethod
    def sanitize_name(name: str) -> str:
        # strip leading characters that are invalid to start
        #  a fieldname (only letters allowed)
        name = name.lstrip("_0123456789")

        # matlab fieldnames allow alphanum and _ after starting characters
        for c in name:
            if not c.isalnum() and c != "_":
                name = name.replace(c, "")
        return name

    def generate_field(self, top_field: str, name: str, value: Any) -> str:
        name = name.replace(f"{top_field}_", "", 1)  # strip top_field from fieldname
        name = self.sanitize_name(name)
        return f"{self.struct_name}.{top_field}.{name} = {value};\n"

    def generate_constant(self, c: Union[ConstantExpr, HID, MID, MT]) -> str:
        if isinstance(c, ConstantExpr):
            prefix = ""
        elif isinstance(c, HID):
            prefix = "HID_"
        elif isinstance(c, MID):
            prefix = "MID_"
        elif isinstance(c, MT):
            prefix = "MT_"
        return self.generate_field(
            "defines", f"{prefix}{self.sanitize_name(c.name)}", c.value
        )

    def generate_constant_string(self, c: ConstantString):
        return self.generate_field("defines", self.sanitize_name(c.name), c.value)

    def generate_host_id(self, hid: HID) -> str:
        return self.generate_field("HID", self.sanitize_name(hid.name), hid.value)

    def generate_module_id(self, mid: MID) -> str:
        return self.generate_field("MID", self.sanitize_name(mid.name), mid.value)

    def generate_msg_type_id(self, mt: MT) -> str:
        return self.generate_field("MT", self.sanitize_name(mt.name), mt.value)

    def generate_type_alias(self, td: TypeAlias) -> str:
        name = self.sanitize_name(td.name)
        type_name = self.sanitize_name(td.type_name)
        if td.type_name in type_map.keys():
            s = type_map[td.type_name]
            return f"{self.struct_name}.typedefs.{name} = {s}(0);\n"

        if td.type_name in self.parser.aliases.keys():
            return f"{self.struct_name}.typedefs.{name} = RTMA.typedefs.{type_name};\n"

        if td.type_name in self.parser.struct_defs.keys():
            return f"{self.struct_name}.typedefs.{name} = RTMA.typedefs.{type_name};\n"

        if td.type_name in self.parser.message_defs.keys():
            return f"{self.struct_name}.MDF.{name} = RTMA.MDF.{type_name};\n"

        raise RuntimeError(f"No type found for alias: {name}")

    def generate_struct(
        self, struct: Union[SDF, MDF], top_field: str = "typedefs"
    ) -> str:
        f = []

        name = self.sanitize_name(struct.name)

        if isinstance(struct, MDF) and len(struct.fields) == 0:
            f.append(f"% {name} (Signal)")
        else:
            f.append(f"% {name}")

        f.append(f"{self.struct_name}.{top_field}.{name} = struct();")

        for field in struct.fields:
            if field.type_name in type_map.keys():
                s = type_map[field.type_name]
                ftype = f"{s}(0)"
            elif field.type_name in self.parser.message_defs.keys():
                ftype = f"{self.struct_name}.MDF.{field.type_name}"
            elif field.type_name in self.parser.struct_defs.keys():
                ftype = f"{self.struct_name}.typedefs.{field.type_name}"
            elif field.type_name in self.parser.aliases.keys():
                ftype = f"{self.struct_name}.typedefs.{field.type_name}"
            else:
                raise RuntimeError(f"Unknown field name {field.name} in {name}")

            if field.length is not None:
                f.append(
                    f"{self.struct_name}.{top_field}.{name}.{field.name} = repmat({ftype}, 1, {field.length});"
                )
            else:
                f.append(
                    f"{self.struct_name}.{top_field}.{name}.{field.name} = {ftype};"
                )

        return "\n".join(f)

    def generate_msg_def(self, mdf: MDF) -> str:
        return self.generate_struct(mdf, "MDF")

    def generate_message_header(self) -> str:
        return f"{self.struct_name}.MESSAGE_HEADER = {self.struct_name}.typedefs.RTMA_MSG_HEADER;\n"

    def generate_hash_id(self, mdf: MDF) -> str:
        return f'{self.struct_name}.hash.{self.sanitize_name(mdf.name)} = "{mdf.hash[:8]}";\n'

    def generate_mex_opcodes(self) -> str:
        # these are copied from MatlabRTMA.h
        prefix = f"{self.struct_name}.mex_opcode"
        op = f"""\
        {prefix}.CONNECT_TO_MMM = 1;
        {prefix}.DISCONNECT_FROM_MMM = 2;
        {prefix}.SUBSCRIBE = 3;
        {prefix}.UNSUBSCRIBE = 4;
        {prefix}.SEND_MESSAGE = 5;
        {prefix}.SEND_SIGNAL = 6;
        {prefix}.READ_MESSAGE_HDR = 7;
        {prefix}.READ_MESSAGE_DATA = 8;
        {prefix}.SET_TIMER = 9;
        {prefix}.SEND_MODULE_READY = 10;
        {prefix}.CANCEL_TIMER = 11;
        {prefix}.PAUSE_SUBSCRIPTION = 12;
        {prefix}.RESUME_SUBSCRIPTION = 13;
        {prefix}.READ_MESSAGE_DD = 14;
        {prefix}.SEND_MESSAGE_DD = 15;
        {prefix}.GET_TIME = 16;
        {prefix}.GET_MODULE_ID = 17;
        """
        return dedent(op)

    def generate_by_MT(self) -> str:
        # append matlab code to generate _by_MT cell arrays from other fields
        prefix = self.struct_name
        s = f"""\n
        % add _by_MT arrays
        mtns = fieldnames({prefix}.MT);
        for idx = 1:length(mtns)
            mtn = mtns{{idx}};
            mt = {prefix}.MT.(mtn) + 1;
            mdf = {prefix}.MDF.(mtn);
            {prefix}.MTN_by_MT{{mt, 1}} = mtn;
            {prefix}.MDF_by_MT{{mt, 1}} = mdf;
        end\n
        """
        return dedent(s)

    def generate_vars(self):
        # do nothing (for now)
        return f"{self.struct_name}.vars = [];\n"

    def generate(self, out_filepath: pathlib.Path):
        with open(out_filepath, mode="w") as f:
            # Version comment
            f.write(
                f"% This message def file was auto-generated by pyrtma.compile version {__version__}\n\n"
            )

            # create RTMA config generator .m function
            f.write(self.generate_fcn_header(out_filepath.stem))

            # Top-Level RTMA object
            f.write("% Top-Level RTMA object\n")
            f.write(self.initialize_struct())
            f.write("\n\n")

            # RTMA.constants
            f.write("% Constants\n")
            obj: Union[ConstantExpr, ConstantString, HID, MID, TypeAlias, SDF, MT, MDF]
            for obj in self.parser.constants.values():
                f.write(self.generate_constant(obj))
            for obj in self.parser.host_ids.values():
                f.write(self.generate_constant(obj))
            for obj in self.parser.module_ids.values():
                f.write(self.generate_constant(obj))
            for obj in self.parser.message_ids.values():
                f.write(self.generate_constant(obj))
            f.write("\n")

            f.write("% String Constants\n")
            for obj in self.parser.string_constants.values():
                f.write(self.generate_constant_string(obj))
            f.write("\n")

            # RTMA.typedefs
            f.write("% Type Aliases\n")
            for obj in self.parser.aliases.values():
                f.write(self.generate_type_alias(obj))
            f.write("\n")

            # RTMA.HID
            f.write("% Host IDs\n")
            for obj in self.parser.host_ids.values():
                f.write(self.generate_host_id(obj))
            f.write("\n")

            # RTMA.MID
            f.write("% Module IDs\n")
            for obj in self.parser.module_ids.values():
                f.write(self.generate_module_id(obj))
            f.write("\n")

            # RTMA.MT
            f.write("% Message Type IDs\n")
            for obj in self.parser.message_ids.values():
                f.write(self.generate_msg_type_id(obj))
            f.write("\n")

            # RTMA.typedefs
            f.write("% Struct Definitions\n")
            for obj in self.parser.struct_defs.values():
                f.write(self.generate_struct(obj))
                f.write("\n\n")

            # RTMA.MDF
            f.write("% Message Definitions\n")
            for obj in self.parser.message_defs.values():
                f.write(self.generate_struct(obj, top_field="MDF"))
                f.write("\n\n")

            # include STRING_DATA core defs for backwards compatibility with quicklogger and message manager output
            f.write("% Manual Definitions - obsolete core defs\n")
            f.write(f"RTMA.MT.MM_ERROR = 83;\n")
            f.write(f"RTMA.MDF.MM_ERROR = 'VARIABLE_LENGTH_ARRAY(int8)';\n")
            f.write(f"RTMA.MT.MM_INFO = 84;\n")
            f.write(f"RTMA.MDF.MM_INFO = 'VARIABLE_LENGTH_ARRAY(int8)';\n")
            f.write(f"RTMA.MT.DEBUG_TEXT = 91;\n")
            f.write(f"RTMA.MDF.DEBUG_TEXT = 'VARIABLE_LENGTH_ARRAY(int8)';\n")

            # RTMA.hash
            f.write("% Message Definition Hashes\n")
            for obj in self.parser.message_defs.values():
                f.write(self.generate_hash_id(obj))
            f.write("\n\n")

            # MESSAGE_HEADER
            f.write(self.generate_message_header())
            f.write("\n\n")

            # mex_opcode
            f.write(self.generate_mex_opcodes())
            f.write("\n\n")

            # vars -- empty struct (for now)
            f.write(self.generate_vars())

            # MTN_by_MT and MDF_by_MT
            f.write(self.generate_by_MT())

            # close function
            f.write(self.generate_fcn_close())



class RTMAMessageError(Exception):
    """Base exception for message errors."""

    pass


class UnknownMessageType(RTMAMessageError):
    """Raised when there is no message definition."""

    pass


class JSONDecodingError(RTMAMessageError):
    """Raised when there is an error decoding a message from json."""

    pass


class InvalidMessageDefinition(RTMAMessageError):
    """Raised when there is message definition is out of sync with sent data."""

    pass


class VersionMismatchWarning(UserWarning):
    """Raised when import message defs were compiled with a different pyrtma version"""

    pass


class ClientError(Exception):
    """Base exception for all Client Errors."""

    pass


class MessageManagerNotFound(ClientError):
    """Raised when unable to connect to message manager."""

    pass


class SocketOptionError(ClientError):
    """Raised when unable to set socket options."""

    pass


class NotConnectedError(ClientError):
    """Raised when the client tries to read/write while not connected."""

    pass


class ConnectionLost(ClientError):
    """Raised when there is a connection error with the server."""

    pass


class AcknowledgementTimeout(ClientError):
    """Raised when client does not receive ack from message manager."""

    pass


class InvalidDestinationModule(ClientError):
    """Raised when client tries to send to an invalid module."""

    pass


class InvalidDestinationHost(ClientError):
    """Raised when client tries to send to an invalid host."""

    pass


class InvalidSubscription(ClientError):
    """Raised when a subscription command is invalid"""

    pass



"""This message def file was auto-generated by pyrtma.compile version 2.3.6"""

import ctypes

import pyrtma
from pyrtma.__version__ import check_compiled_version, __version__
from packaging.version import parse as _ver_parse
from typing import ClassVar

from pyrtma.message_base import MessageBase, MessageMeta
from pyrtma.message_data import MessageData
from pyrtma.validators import (
    Int8,
    Int16,
    Int32,
    Int64,
    Uint8,
    Uint16,
    Uint32,
    Uint64,
    Float,
    Double,
    Struct,
    IntArray,
    FloatArray,
    StructArray,
    Char,
    String,
    Byte,
    ByteArray,
)

if _ver_parse(__version__) >= _ver_parse("2.3.0"):
    from pyrtma.context import _update_context, get_context


COMPILED_PYRTMA_VERSION: str = "2.3.6"
check_compiled_version(COMPILED_PYRTMA_VERSION)

# Constants
MAX_DATA_SETS: int = 6
MAX_LOGGER_FILENAME_LENGTH: int = 256
MAX_MODULES: int = 200
DYN_MOD_ID_START: int = 100
MAX_HOSTS: int = 5
MAX_MESSAGE_TYPES: int = 10000
MAX_RTMA_MSG_TYPE: int = 99
MAX_RTMA_MODULE_ID: int = 9
MAX_CONTIGUOUS_MESSAGE_DATA: int = 9000
ALL_MESSAGE_TYPES: int = 2147483647
MAX_ACTIVE_CLIENTS: int = 256
MAX_SUBSCRIBERS: int = 256
MAX_SUBS: int = 256
MAX_NAME_LEN: int = 32
MAX_LOG_LENGTH: int = 1024
MESSAGE_TRAFFIC_SIZE: int = 64
MAX_MESSAGE_SIZE: int = 65535

# String Constants

# Type Aliases
MODULE_ID = ctypes.c_int16
HOST_ID = ctypes.c_int16
MSG_TYPE = ctypes.c_int32
MSG_COUNT = ctypes.c_int32

# Host IDs
LOCAL_HOST: int = 0
ALL_HOSTS: int = 32767

# Module IDs
MID_MESSAGE_MANAGER: int = 0
MID_DATA_LOGGER: int = 4
MID_QUICK_LOGGER: int = 5

# Message Type IDs
MT_DATA_LOGGER_STATUS: int = 62
MT_DATA_LOGGER_STATUS_REQUEST: int = 63
MT_ADD_DATA_COLLECTION: int = 64
MT_REMOVE_DATA_COLLECTION: int = 65
MT_ADD_DATA_SET: int = 66
MT_REMOVE_DATA_SET: int = 67
MT_DATA_COLLECTION_CONFIG_REQUEST: int = 68
MT_DATA_COLLECTION_CONFIG: int = 69
MT_DATA_COLLECTION_STARTED: int = 70
MT_DATA_COLLECTION_STOPPED: int = 71
MT_DATA_COLLECTION_SAVED: int = 72
MT_DATA_LOGGER_START: int = 73
MT_DATA_LOGGER_STOP: int = 74
MT_DATA_LOGGER_PAUSE: int = 75
MT_DATA_LOGGER_RESUME: int = 76
MT_DATA_LOGGER_RESET: int = 77
MT_DATA_LOGGER_ERROR: int = 78
MT_DATA_LOGGER_METADATA_UPDATE: int = 79
MT_DATA_LOGGER_METADATA_REQUEST: int = 87
MT_DATA_LOGGER_METADATA: int = 88
MT_DATA_LOG_TEST_2048: int = 89
MT_LM_STATUS: int = 54
MT_LM_EXIT: int = 55
MT_SAVE_MESSAGE_LOG: int = 56
MT_MESSAGE_LOG_SAVED: int = 57
MT_PAUSE_MESSAGE_LOGGING: int = 58
MT_RESUME_MESSAGE_LOGGING: int = 59
MT_RESET_MESSAGE_LOG: int = 60
MT_DUMP_MESSAGE_LOG: int = 61
MT_LM_READY: int = 96
MT_EXIT: int = 0
MT_KILL: int = 1
MT_ACKNOWLEDGE: int = 2
MT_CONNECT_V2: int = 4
MT_FAIL_SUBSCRIBE: int = 6
MT_FAILED_MESSAGE: int = 8
MT_CONNECT: int = 13
MT_DISCONNECT: int = 14
MT_SUBSCRIBE: int = 15
MT_UNSUBSCRIBE: int = 16
MT_MODULE_READY: int = 26
MT_MESSAGE_TRAFFIC: int = 30
MT_ACTIVE_CLIENTS: int = 31
MT_CLIENT_INFO: int = 32
MT_CLIENT_CLOSED: int = 33
MT_CLIENT_SET_NAME: int = 34
MT_RTMA_LOG: int = 40
MT_RTMA_LOG_CRITICAL: int = 41
MT_RTMA_LOG_ERROR: int = 42
MT_RTMA_LOG_WARNING: int = 43
MT_RTMA_LOG_INFO: int = 44
MT_RTMA_LOG_DEBUG: int = 45
MT_TIMING_MESSAGE: int = 80
MT_FORCE_DISCONNECT: int = 82
MT_PAUSE_SUBSCRIPTION: int = 85
MT_RESUME_SUBSCRIPTION: int = 86


# Struct Definitions
class DATA_SET(MessageBase, metaclass=MessageMeta):
    type_name: ClassVar[str] = "DATA_SET"
    type_hash: ClassVar[int] = 0xB267DE7B
    type_size: ClassVar[int] = 452
    type_source: ClassVar[str] = "data_logger.yaml"
    type_def: ClassVar[str] = (
        "'DATA_SET:\n  fields:\n    name: char[32]\n    sub_dir_fmt: char[128]\n    file_name_fmt: char[128]\n    formatter: char[32]\n    subdivide_interval: int32\n    msg_types: int32[32]'"
    )

    name: String = String(32)
    sub_dir_fmt: String = String(128)
    file_name_fmt: String = String(128)
    formatter: String = String(32)
    subdivide_interval: Int32 = Int32()
    msg_types: IntArray[Int32] = IntArray(Int32, 32)


class DATA_COLLECTION(MessageBase, metaclass=MessageMeta):
    type_name: ClassVar[str] = "DATA_COLLECTION"
    type_hash: ClassVar[int] = 0x9FA9C51A
    type_size: ClassVar[int] = 3132
    type_source: ClassVar[str] = "data_logger.yaml"
    type_def: ClassVar[str] = (
        "'DATA_COLLECTION:\n  fields:\n    name: char[32]\n    base_path: char[256]\n    dir_fmt: char[128]\n    num_data_sets: int16\n    unused: int16\n    data_sets: DATA_SET[MAX_DATA_SETS]'"
    )

    name: String = String(32)
    base_path: String = String(256)
    dir_fmt: String = String(128)
    num_data_sets: Int16 = Int16()
    unused: Int16 = Int16()
    data_sets: StructArray[DATA_SET] = StructArray(DATA_SET, 6)


class DATA_SET_INFO(MessageBase, metaclass=MessageMeta):
    type_name: ClassVar[str] = "DATA_SET_INFO"
    type_hash: ClassVar[int] = 0x65972A16
    type_size: ClassVar[int] = 452
    type_source: ClassVar[str] = "data_logger.yaml"
    type_def: ClassVar[str] = (
        "'DATA_SET_INFO:\n  fields:\n    name: char[32]\n    formatter: char[32]\n    save_path: char[256]\n    subdivide_interval: int32\n    msg_types: int32[32]'"
    )

    name: String = String(32)
    formatter: String = String(32)
    save_path: String = String(256)
    subdivide_interval: Int32 = Int32()
    msg_types: IntArray[Int32] = IntArray(Int32, 32)


class DATA_COLLECTION_INFO(MessageBase, metaclass=MessageMeta):
    type_name: ClassVar[str] = "DATA_COLLECTION_INFO"
    type_hash: ClassVar[int] = 0x8FAF0CA3
    type_size: ClassVar[int] = 3004
    type_source: ClassVar[str] = "data_logger.yaml"
    type_def: ClassVar[str] = (
        "'DATA_COLLECTION_INFO:\n  fields:\n    name: char[32]\n    save_path: char[256]\n    num_data_sets: int16\n    unused: int16\n    data_sets: DATA_SET_INFO[MAX_DATA_SETS]'"
    )

    name: String = String(32)
    save_path: String = String(256)
    num_data_sets: Int16 = Int16()
    unused: Int16 = Int16()
    data_sets: StructArray[DATA_SET_INFO] = StructArray(DATA_SET_INFO, 6)


class RTMA_MSG_HEADER(MessageBase, metaclass=MessageMeta):
    type_name: ClassVar[str] = "RTMA_MSG_HEADER"
    type_hash: ClassVar[int] = 0x35CC5A39
    type_size: ClassVar[int] = 48
    type_source: ClassVar[str] = "core_defs.yaml"
    type_def: ClassVar[str] = (
        "'RTMA_MSG_HEADER:\n  fields:\n    msg_type: MSG_TYPE\n    msg_count: MSG_COUNT\n    send_time: double\n    recv_time: double\n    src_host_id: HOST_ID\n    src_mod_id: MODULE_ID\n    dest_host_id: HOST_ID\n    dest_mod_id: MODULE_ID\n    num_data_bytes: int32\n    remaining_bytes: int32\n    is_dynamic: int32\n    reserved: uint32'"
    )

    msg_type: Int32 = Int32()
    msg_count: Int32 = Int32()
    send_time: Double = Double()
    recv_time: Double = Double()
    src_host_id: Int16 = Int16()
    src_mod_id: Int16 = Int16()
    dest_host_id: Int16 = Int16()
    dest_mod_id: Int16 = Int16()
    num_data_bytes: Int32 = Int32()
    remaining_bytes: Int32 = Int32()
    is_dynamic: Int32 = Int32()
    reserved: Uint32 = Uint32()


# Message Definitions
@pyrtma.message_def
class MDF_DATA_LOGGER_STATUS(MessageData, metaclass=MessageMeta):
    type_id: ClassVar[int] = 62
    type_name: ClassVar[str] = "DATA_LOGGER_STATUS"
    type_hash: ClassVar[int] = 0x9E922E93
    type_size: ClassVar[int] = 24
    type_source: ClassVar[str] = "data_logger.yaml"
    type_def: ClassVar[str] = (
        "'DATA_LOGGER_STATUS:\n  id: 62\n  fields:\n    timestamp: double\n    elapsed_time: double\n    is_recording: int32\n    is_paused: int32'"
    )

    timestamp: Double = Double()
    elapsed_time: Double = Double()
    is_recording: Int32 = Int32()
    is_paused: Int32 = Int32()


@pyrtma.message_def
class MDF_DATA_LOGGER_STATUS_REQUEST(MessageData, metaclass=MessageMeta):
    type_id: ClassVar[int] = 63
    type_name: ClassVar[str] = "DATA_LOGGER_STATUS_REQUEST"
    type_hash: ClassVar[int] = 0x168B0652
    type_size: ClassVar[int] = 0
    type_source: ClassVar[str] = "data_logger.yaml"
    type_def: ClassVar[str] = "'DATA_LOGGER_STATUS_REQUEST:\n  id: 63\n  fields: null'"


@pyrtma.message_def
class MDF_ADD_DATA_COLLECTION(MessageData, metaclass=MessageMeta):
    type_id: ClassVar[int] = 64
    type_name: ClassVar[str] = "ADD_DATA_COLLECTION"
    type_hash: ClassVar[int] = 0xA50C4BC8
    type_size: ClassVar[int] = 3132
    type_source: ClassVar[str] = "data_logger.yaml"
    type_def: ClassVar[str] = (
        "'ADD_DATA_COLLECTION:\n  id: 64\n  fields:\n    collection: DATA_COLLECTION'"
    )

    collection: Struct[DATA_COLLECTION] = Struct(DATA_COLLECTION)


@pyrtma.message_def
class MDF_REMOVE_DATA_COLLECTION(MessageData, metaclass=MessageMeta):
    type_id: ClassVar[int] = 65
    type_name: ClassVar[str] = "REMOVE_DATA_COLLECTION"
    type_hash: ClassVar[int] = 0x5430DCF2
    type_size: ClassVar[int] = 32
    type_source: ClassVar[str] = "data_logger.yaml"
    type_def: ClassVar[str] = (
        "'REMOVE_DATA_COLLECTION:\n  id: 65\n  fields:\n    collection_name: char[32]'"
    )

    collection_name: String = String(32)


@pyrtma.message_def
class MDF_ADD_DATA_SET(MessageData, metaclass=MessageMeta):
    type_id: ClassVar[int] = 66
    type_name: ClassVar[str] = "ADD_DATA_SET"
    type_hash: ClassVar[int] = 0xB368D74F
    type_size: ClassVar[int] = 484
    type_source: ClassVar[str] = "data_logger.yaml"
    type_def: ClassVar[str] = (
        "'ADD_DATA_SET:\n  id: 66\n  fields:\n    collection_name: char[32]\n    data_set: DATA_SET'"
    )

    collection_name: String = String(32)
    data_set: Struct[DATA_SET] = Struct(DATA_SET)


@pyrtma.message_def
class MDF_REMOVE_DATA_SET(MessageData, metaclass=MessageMeta):
    type_id: ClassVar[int] = 67
    type_name: ClassVar[str] = "REMOVE_DATA_SET"
    type_hash: ClassVar[int] = 0x0BC39C04
    type_size: ClassVar[int] = 64
    type_source: ClassVar[str] = "data_logger.yaml"
    type_def: ClassVar[str] = (
        "'REMOVE_DATA_SET:\n  id: 67\n  fields:\n    collection_name: char[32]\n    name: char[32]'"
    )

    collection_name: String = String(32)
    name: String = String(32)


@pyrtma.message_def
class MDF_DATA_COLLECTION_CONFIG_REQUEST(MessageData, metaclass=MessageMeta):
    type_id: ClassVar[int] = 68
    type_name: ClassVar[str] = "DATA_COLLECTION_CONFIG_REQUEST"
    type_hash: ClassVar[int] = 0x1779D826
    type_size: ClassVar[int] = 0
    type_source: ClassVar[str] = "data_logger.yaml"
    type_def: ClassVar[str] = (
        "'DATA_COLLECTION_CONFIG_REQUEST:\n  id: 68\n  fields: null'"
    )


@pyrtma.message_def
class MDF_DATA_COLLECTION_CONFIG(MessageData, metaclass=MessageMeta):
    type_id: ClassVar[int] = 69
    type_name: ClassVar[str] = "DATA_COLLECTION_CONFIG"
    type_hash: ClassVar[int] = 0xC08D3576
    type_size: ClassVar[int] = 3132
    type_source: ClassVar[str] = "data_logger.yaml"
    type_def: ClassVar[str] = (
        "'DATA_COLLECTION_CONFIG:\n  id: 69\n  fields:\n    collection: DATA_COLLECTION'"
    )

    collection: Struct[DATA_COLLECTION] = Struct(DATA_COLLECTION)


@pyrtma.message_def
class MDF_DATA_COLLECTION_STARTED(MessageData, metaclass=MessageMeta):
    type_id: ClassVar[int] = 70
    type_name: ClassVar[str] = "DATA_COLLECTION_STARTED"
    type_hash: ClassVar[int] = 0x415737CA
    type_size: ClassVar[int] = 3004
    type_source: ClassVar[str] = "data_logger.yaml"
    type_def: ClassVar[str] = (
        "'DATA_COLLECTION_STARTED:\n  id: 70\n  fields:\n    collection: DATA_COLLECTION_INFO'"
    )

    collection: Struct[DATA_COLLECTION_INFO] = Struct(DATA_COLLECTION_INFO)


@pyrtma.message_def
class MDF_DATA_COLLECTION_STOPPED(MessageData, metaclass=MessageMeta):
    type_id: ClassVar[int] = 71
    type_name: ClassVar[str] = "DATA_COLLECTION_STOPPED"
    type_hash: ClassVar[int] = 0xDA1AF3AE
    type_size: ClassVar[int] = 3004
    type_source: ClassVar[str] = "data_logger.yaml"
    type_def: ClassVar[str] = (
        "'DATA_COLLECTION_STOPPED:\n  id: 71\n  fields:\n    collection: DATA_COLLECTION_INFO'"
    )

    collection: Struct[DATA_COLLECTION_INFO] = Struct(DATA_COLLECTION_INFO)


@pyrtma.message_def
class MDF_DATA_COLLECTION_SAVED(MessageData, metaclass=MessageMeta):
    type_id: ClassVar[int] = 72
    type_name: ClassVar[str] = "DATA_COLLECTION_SAVED"
    type_hash: ClassVar[int] = 0x74528518
    type_size: ClassVar[int] = 0
    type_source: ClassVar[str] = "data_logger.yaml"
    type_def: ClassVar[str] = "'DATA_COLLECTION_SAVED:\n  id: 72\n  fields: null'"


@pyrtma.message_def
class MDF_DATA_LOGGER_START(MessageData, metaclass=MessageMeta):
    type_id: ClassVar[int] = 73
    type_name: ClassVar[str] = "DATA_LOGGER_START"
    type_hash: ClassVar[int] = 0x982D6942
    type_size: ClassVar[int] = 0
    type_source: ClassVar[str] = "data_logger.yaml"
    type_def: ClassVar[str] = "'DATA_LOGGER_START:\n  id: 73\n  fields: null'"


@pyrtma.message_def
class MDF_DATA_LOGGER_STOP(MessageData, metaclass=MessageMeta):
    type_id: ClassVar[int] = 74
    type_name: ClassVar[str] = "DATA_LOGGER_STOP"
    type_hash: ClassVar[int] = 0xA5B66B24
    type_size: ClassVar[int] = 0
    type_source: ClassVar[str] = "data_logger.yaml"
    type_def: ClassVar[str] = "'DATA_LOGGER_STOP:\n  id: 74\n  fields: null'"


@pyrtma.message_def
class MDF_DATA_LOGGER_PAUSE(MessageData, metaclass=MessageMeta):
    type_id: ClassVar[int] = 75
    type_name: ClassVar[str] = "DATA_LOGGER_PAUSE"
    type_hash: ClassVar[int] = 0x9A556835
    type_size: ClassVar[int] = 0
    type_source: ClassVar[str] = "data_logger.yaml"
    type_def: ClassVar[str] = "'DATA_LOGGER_PAUSE:\n  id: 75\n  fields: null'"


@pyrtma.message_def
class MDF_DATA_LOGGER_RESUME(MessageData, metaclass=MessageMeta):
    type_id: ClassVar[int] = 76
    type_name: ClassVar[str] = "DATA_LOGGER_RESUME"
    type_hash: ClassVar[int] = 0x18811751
    type_size: ClassVar[int] = 0
    type_source: ClassVar[str] = "data_logger.yaml"
    type_def: ClassVar[str] = "'DATA_LOGGER_RESUME:\n  id: 76\n  fields: null'"


@pyrtma.message_def
class MDF_DATA_LOGGER_RESET(MessageData, metaclass=MessageMeta):
    type_id: ClassVar[int] = 77
    type_name: ClassVar[str] = "DATA_LOGGER_RESET"
    type_hash: ClassVar[int] = 0x1EB32BEB
    type_size: ClassVar[int] = 0
    type_source: ClassVar[str] = "data_logger.yaml"
    type_def: ClassVar[str] = "'DATA_LOGGER_RESET:\n  id: 77\n  fields: null'"


@pyrtma.message_def
class MDF_DATA_LOGGER_ERROR(MessageData, metaclass=MessageMeta):
    type_id: ClassVar[int] = 78
    type_name: ClassVar[str] = "DATA_LOGGER_ERROR"
    type_hash: ClassVar[int] = 0xAD3B4A67
    type_size: ClassVar[int] = 512
    type_source: ClassVar[str] = "data_logger.yaml"
    type_def: ClassVar[str] = (
        "'DATA_LOGGER_ERROR:\n  id: 78\n  fields:\n    msg: char[512]'"
    )

    msg: String = String(512)


@pyrtma.message_def
class MDF_DATA_LOGGER_METADATA_UPDATE(MessageData, metaclass=MessageMeta):
    type_id: ClassVar[int] = 79
    type_name: ClassVar[str] = "DATA_LOGGER_METADATA_UPDATE"
    type_hash: ClassVar[int] = 0x3C673526
    type_size: ClassVar[int] = 512
    type_source: ClassVar[str] = "data_logger.yaml"
    type_def: ClassVar[str] = (
        "'DATA_LOGGER_METADATA_UPDATE:\n  id: 79\n  fields:\n    json: char[512]'"
    )

    json: String = String(512)


@pyrtma.message_def
class MDF_DATA_LOGGER_METADATA_REQUEST(MessageData, metaclass=MessageMeta):
    type_id: ClassVar[int] = 87
    type_name: ClassVar[str] = "DATA_LOGGER_METADATA_REQUEST"
    type_hash: ClassVar[int] = 0x2D068003
    type_size: ClassVar[int] = 0
    type_source: ClassVar[str] = "data_logger.yaml"
    type_def: ClassVar[str] = (
        "'DATA_LOGGER_METADATA_REQUEST:\n  id: 87\n  fields: null'"
    )


@pyrtma.message_def
class MDF_DATA_LOGGER_METADATA(MessageData, metaclass=MessageMeta):
    type_id: ClassVar[int] = 88
    type_name: ClassVar[str] = "DATA_LOGGER_METADATA"
    type_hash: ClassVar[int] = 0x3852789B
    type_size: ClassVar[int] = 1024
    type_source: ClassVar[str] = "data_logger.yaml"
    type_def: ClassVar[str] = (
        "'DATA_LOGGER_METADATA:\n  id: 88\n  fields:\n    json: char[1024]'"
    )

    json: String = String(1024)


@pyrtma.message_def
class MDF_DATA_LOG_TEST_2048(MessageData, metaclass=MessageMeta):
    type_id: ClassVar[int] = 89
    type_name: ClassVar[str] = "DATA_LOG_TEST_2048"
    type_hash: ClassVar[int] = 0x651D3C98
    type_size: ClassVar[int] = 2048
    type_source: ClassVar[str] = "data_logger.yaml"
    type_def: ClassVar[str] = (
        "'DATA_LOG_TEST_2048:\n  id: 89\n  fields:\n    raw: char[2048]'"
    )

    raw: String = String(2048)


@pyrtma.message_def
class MDF_LM_STATUS(MessageData, metaclass=MessageMeta):
    type_id: ClassVar[int] = 54
    type_name: ClassVar[str] = "LM_STATUS"
    type_hash: ClassVar[int] = 0x2DA5B6A1
    type_size: ClassVar[int] = 32
    type_source: ClassVar[str] = "quick_logger.yaml"
    type_def: ClassVar[str] = (
        "'LM_STATUS:\n  id: 54\n  fields:\n    is_logging: uint32\n    max_msgs: uint32\n    hdr_bufsz: uint32\n    data_bufsz: uint32\n    msg_count: uint32\n    hdr_total: uint32\n    data_total: uint32\n    ofs_total: uint32'"
    )

    is_logging: Uint32 = Uint32()
    max_msgs: Uint32 = Uint32()
    hdr_bufsz: Uint32 = Uint32()
    data_bufsz: Uint32 = Uint32()
    msg_count: Uint32 = Uint32()
    hdr_total: Uint32 = Uint32()
    data_total: Uint32 = Uint32()
    ofs_total: Uint32 = Uint32()


@pyrtma.message_def
class MDF_LM_EXIT(MessageData, metaclass=MessageMeta):
    type_id: ClassVar[int] = 55
    type_name: ClassVar[str] = "LM_EXIT"
    type_hash: ClassVar[int] = 0x35DD547B
    type_size: ClassVar[int] = 0
    type_source: ClassVar[str] = "quick_logger.yaml"
    type_def: ClassVar[str] = "'LM_EXIT:\n  id: 55\n  fields: null'"


@pyrtma.message_def
class MDF_SAVE_MESSAGE_LOG(MessageData, metaclass=MessageMeta):
    type_id: ClassVar[int] = 56
    type_name: ClassVar[str] = "SAVE_MESSAGE_LOG"
    type_hash: ClassVar[int] = 0xF0D7A3A9
    type_size: ClassVar[int] = 260
    type_source: ClassVar[str] = "quick_logger.yaml"
    type_def: ClassVar[str] = (
        "'SAVE_MESSAGE_LOG:\n  id: 56\n  fields:\n    pathname: char[MAX_LOGGER_FILENAME_LENGTH]\n    pathname_length: int32'"
    )

    pathname: String = String(256)
    pathname_length: Int32 = Int32()


@pyrtma.message_def
class MDF_MESSAGE_LOG_SAVED(MessageData, metaclass=MessageMeta):
    type_id: ClassVar[int] = 57
    type_name: ClassVar[str] = "MESSAGE_LOG_SAVED"
    type_hash: ClassVar[int] = 0x66E84AE5
    type_size: ClassVar[int] = 0
    type_source: ClassVar[str] = "quick_logger.yaml"
    type_def: ClassVar[str] = "'MESSAGE_LOG_SAVED:\n  id: 57\n  fields: null'"


@pyrtma.message_def
class MDF_PAUSE_MESSAGE_LOGGING(MessageData, metaclass=MessageMeta):
    type_id: ClassVar[int] = 58
    type_name: ClassVar[str] = "PAUSE_MESSAGE_LOGGING"
    type_hash: ClassVar[int] = 0x20C1E922
    type_size: ClassVar[int] = 0
    type_source: ClassVar[str] = "quick_logger.yaml"
    type_def: ClassVar[str] = "'PAUSE_MESSAGE_LOGGING:\n  id: 58\n  fields: null'"


@pyrtma.message_def
class MDF_RESUME_MESSAGE_LOGGING(MessageData, metaclass=MessageMeta):
    type_id: ClassVar[int] = 59
    type_name: ClassVar[str] = "RESUME_MESSAGE_LOGGING"
    type_hash: ClassVar[int] = 0x0D1A3E77
    type_size: ClassVar[int] = 0
    type_source: ClassVar[str] = "quick_logger.yaml"
    type_def: ClassVar[str] = "'RESUME_MESSAGE_LOGGING:\n  id: 59\n  fields: null'"


@pyrtma.message_def
class MDF_RESET_MESSAGE_LOG(MessageData, metaclass=MessageMeta):
    type_id: ClassVar[int] = 60
    type_name: ClassVar[str] = "RESET_MESSAGE_LOG"
    type_hash: ClassVar[int] = 0x68EC4AAB
    type_size: ClassVar[int] = 0
    type_source: ClassVar[str] = "quick_logger.yaml"
    type_def: ClassVar[str] = "'RESET_MESSAGE_LOG:\n  id: 60\n  fields: null'"


@pyrtma.message_def
class MDF_DUMP_MESSAGE_LOG(MessageData, metaclass=MessageMeta):
    type_id: ClassVar[int] = 61
    type_name: ClassVar[str] = "DUMP_MESSAGE_LOG"
    type_hash: ClassVar[int] = 0xF9D7E2BF
    type_size: ClassVar[int] = 0
    type_source: ClassVar[str] = "quick_logger.yaml"
    type_def: ClassVar[str] = "'DUMP_MESSAGE_LOG:\n  id: 61\n  fields: null'"


@pyrtma.message_def
class MDF_LM_READY(MessageData, metaclass=MessageMeta):
    type_id: ClassVar[int] = 96
    type_name: ClassVar[str] = "LM_READY"
    type_hash: ClassVar[int] = 0x4863B960
    type_size: ClassVar[int] = 0
    type_source: ClassVar[str] = "quick_logger.yaml"
    type_def: ClassVar[str] = "'LM_READY:\n  id: 96\n  fields: null'"


@pyrtma.message_def
class MDF_EXIT(MessageData, metaclass=MessageMeta):
    type_id: ClassVar[int] = 0
    type_name: ClassVar[str] = "EXIT"
    type_hash: ClassVar[int] = 0x095E0546
    type_size: ClassVar[int] = 0
    type_source: ClassVar[str] = "core_defs.yaml"
    type_def: ClassVar[str] = "'EXIT:\n  id: 0\n  fields: null'"


@pyrtma.message_def
class MDF_KILL(MessageData, metaclass=MessageMeta):
    type_id: ClassVar[int] = 1
    type_name: ClassVar[str] = "KILL"
    type_hash: ClassVar[int] = 0x82FC702D
    type_size: ClassVar[int] = 0
    type_source: ClassVar[str] = "core_defs.yaml"
    type_def: ClassVar[str] = "'KILL:\n  id: 1\n  fields: null'"


@pyrtma.message_def
class MDF_ACKNOWLEDGE(MessageData, metaclass=MessageMeta):
    type_id: ClassVar[int] = 2
    type_name: ClassVar[str] = "ACKNOWLEDGE"
    type_hash: ClassVar[int] = 0xB725B581
    type_size: ClassVar[int] = 0
    type_source: ClassVar[str] = "core_defs.yaml"
    type_def: ClassVar[str] = "'ACKNOWLEDGE:\n  id: 2\n  fields: null'"


@pyrtma.message_def
class MDF_CONNECT_V2(MessageData, metaclass=MessageMeta):
    type_id: ClassVar[int] = 4
    type_name: ClassVar[str] = "CONNECT_V2"
    type_hash: ClassVar[int] = 0x5CF13682
    type_size: ClassVar[int] = 44
    type_source: ClassVar[str] = "core_defs.yaml"
    type_def: ClassVar[str] = (
        "'CONNECT_V2:\n  id: 4\n  fields:\n    logger_status: int16\n    daemon_status: int16\n    allow_multiple: int16\n    mod_id: MODULE_ID\n    pid: int32\n    name: char[MAX_NAME_LEN]'"
    )

    logger_status: Int16 = Int16()
    daemon_status: Int16 = Int16()
    allow_multiple: Int16 = Int16()
    mod_id: Int16 = Int16()
    pid: Int32 = Int32()
    name: String = String(32)


@pyrtma.message_def
class MDF_FAIL_SUBSCRIBE(MessageData, metaclass=MessageMeta):
    type_id: ClassVar[int] = 6
    type_name: ClassVar[str] = "FAIL_SUBSCRIBE"
    type_hash: ClassVar[int] = 0x694BF3DB
    type_size: ClassVar[int] = 8
    type_source: ClassVar[str] = "core_defs.yaml"
    type_def: ClassVar[str] = (
        "'FAIL_SUBSCRIBE:\n  id: 6\n  fields:\n    mod_id: MODULE_ID\n    reserved: int16\n    msg_type: MSG_TYPE'"
    )

    mod_id: Int16 = Int16()
    reserved: Int16 = Int16()
    msg_type: Int32 = Int32()


@pyrtma.message_def
class MDF_FAILED_MESSAGE(MessageData, metaclass=MessageMeta):
    type_id: ClassVar[int] = 8
    type_name: ClassVar[str] = "FAILED_MESSAGE"
    type_hash: ClassVar[int] = 0x4EAFB837
    type_size: ClassVar[int] = 64
    type_source: ClassVar[str] = "core_defs.yaml"
    type_def: ClassVar[str] = (
        "'FAILED_MESSAGE:\n  id: 8\n  fields:\n    dest_mod_id: MODULE_ID\n    reserved: int16[3]\n    time_of_failure: double\n    msg_header: RTMA_MSG_HEADER'"
    )

    dest_mod_id: Int16 = Int16()
    reserved: IntArray[Int16] = IntArray(Int16, 3)
    time_of_failure: Double = Double()
    msg_header: Struct[RTMA_MSG_HEADER] = Struct(RTMA_MSG_HEADER)


@pyrtma.message_def
class MDF_CONNECT(MessageData, metaclass=MessageMeta):
    type_id: ClassVar[int] = 13
    type_name: ClassVar[str] = "CONNECT"
    type_hash: ClassVar[int] = 0xF4FDCFF3
    type_size: ClassVar[int] = 4
    type_source: ClassVar[str] = "core_defs.yaml"
    type_def: ClassVar[str] = (
        "'CONNECT:\n  id: 13\n  fields:\n    logger_status: int16\n    daemon_status: int16'"
    )

    logger_status: Int16 = Int16()
    daemon_status: Int16 = Int16()


@pyrtma.message_def
class MDF_DISCONNECT(MessageData, metaclass=MessageMeta):
    type_id: ClassVar[int] = 14
    type_name: ClassVar[str] = "DISCONNECT"
    type_hash: ClassVar[int] = 0xD0126BF9
    type_size: ClassVar[int] = 0
    type_source: ClassVar[str] = "core_defs.yaml"
    type_def: ClassVar[str] = "'DISCONNECT:\n  id: 14\n  fields: null'"


@pyrtma.message_def
class MDF_SUBSCRIBE(MessageData, metaclass=MessageMeta):
    type_id: ClassVar[int] = 15
    type_name: ClassVar[str] = "SUBSCRIBE"
    type_hash: ClassVar[int] = 0xF5B437C8
    type_size: ClassVar[int] = 4
    type_source: ClassVar[str] = "core_defs.yaml"
    type_def: ClassVar[str] = (
        "'SUBSCRIBE:\n  id: 15\n  fields:\n    msg_type: MSG_TYPE'"
    )

    msg_type: Int32 = Int32()


@pyrtma.message_def
class MDF_UNSUBSCRIBE(MessageData, metaclass=MessageMeta):
    type_id: ClassVar[int] = 16
    type_name: ClassVar[str] = "UNSUBSCRIBE"
    type_hash: ClassVar[int] = 0x193FB9E0
    type_size: ClassVar[int] = 4
    type_source: ClassVar[str] = "core_defs.yaml"
    type_def: ClassVar[str] = (
        "'UNSUBSCRIBE:\n  id: 16\n  fields:\n    msg_type: MSG_TYPE'"
    )

    msg_type: Int32 = Int32()


@pyrtma.message_def
class MDF_MODULE_READY(MessageData, metaclass=MessageMeta):
    type_id: ClassVar[int] = 26
    type_name: ClassVar[str] = "MODULE_READY"
    type_hash: ClassVar[int] = 0xFD0E0311
    type_size: ClassVar[int] = 4
    type_source: ClassVar[str] = "core_defs.yaml"
    type_def: ClassVar[str] = "'MODULE_READY:\n  id: 26\n  fields:\n    pid: int32'"

    pid: Int32 = Int32()


@pyrtma.message_def
class MDF_MESSAGE_TRAFFIC(MessageData, metaclass=MessageMeta):
    type_id: ClassVar[int] = 30
    type_name: ClassVar[str] = "MESSAGE_TRAFFIC"
    type_hash: ClassVar[int] = 0xA5D05FF5
    type_size: ClassVar[int] = 408
    type_source: ClassVar[str] = "core_defs.yaml"
    type_def: ClassVar[str] = (
        "'MESSAGE_TRAFFIC:\n  id: 30\n  fields:\n    seqno: uint32\n    sub_seqno: uint32\n    start_timestamp: double\n    end_timestamp: double\n    msg_type: MSG_TYPE[MESSAGE_TRAFFIC_SIZE]\n    msg_count: uint16[MESSAGE_TRAFFIC_SIZE]'"
    )

    seqno: Uint32 = Uint32()
    sub_seqno: Uint32 = Uint32()
    start_timestamp: Double = Double()
    end_timestamp: Double = Double()
    msg_type: IntArray[Int32] = IntArray(Int32, 64)
    msg_count: IntArray[Uint16] = IntArray(Uint16, 64)


@pyrtma.message_def
class MDF_ACTIVE_CLIENTS(MessageData, metaclass=MessageMeta):
    type_id: ClassVar[int] = 31
    type_name: ClassVar[str] = "ACTIVE_CLIENTS"
    type_hash: ClassVar[int] = 0xFC42B3AF
    type_size: ClassVar[int] = 1552
    type_source: ClassVar[str] = "core_defs.yaml"
    type_def: ClassVar[str] = (
        "'ACTIVE_CLIENTS:\n  id: 31\n  fields:\n    timestamp: double\n    num_clients: int16\n    padding: int16\n    reserved: int32\n    client_mod_id: MODULE_ID[MAX_ACTIVE_CLIENTS]\n    client_pid: int32[MAX_ACTIVE_CLIENTS]'"
    )

    timestamp: Double = Double()
    num_clients: Int16 = Int16()
    padding: Int16 = Int16()
    reserved: Int32 = Int32()
    client_mod_id: IntArray[Int16] = IntArray(Int16, 256)
    client_pid: IntArray[Int32] = IntArray(Int32, 256)


@pyrtma.message_def
class MDF_CLIENT_INFO(MessageData, metaclass=MessageMeta):
    type_id: ClassVar[int] = 32
    type_name: ClassVar[str] = "CLIENT_INFO"
    type_hash: ClassVar[int] = 0x401076E7
    type_size: ClassVar[int] = 80
    type_source: ClassVar[str] = "core_defs.yaml"
    type_def: ClassVar[str] = (
        "'CLIENT_INFO:\n  id: 32\n  fields:\n    addr: char[32]\n    uid: int32\n    pid: int32\n    mod_id: MODULE_ID\n    is_logger: int16\n    is_unique: int16\n    port: uint16\n    name: char[MAX_NAME_LEN]'"
    )

    addr: String = String(32)
    uid: Int32 = Int32()
    pid: Int32 = Int32()
    mod_id: Int16 = Int16()
    is_logger: Int16 = Int16()
    is_unique: Int16 = Int16()
    port: Uint16 = Uint16()
    name: String = String(32)


@pyrtma.message_def
class MDF_CLIENT_CLOSED(MessageData, metaclass=MessageMeta):
    type_id: ClassVar[int] = 33
    type_name: ClassVar[str] = "CLIENT_CLOSED"
    type_hash: ClassVar[int] = 0x087F4249
    type_size: ClassVar[int] = 80
    type_source: ClassVar[str] = "core_defs.yaml"
    type_def: ClassVar[str] = (
        "'CLIENT_CLOSED:\n  id: 33\n  fields:\n    addr: char[32]\n    uid: int32\n    pid: int32\n    mod_id: MODULE_ID\n    is_logger: int16\n    is_unique: int16\n    port: uint16\n    name: char[MAX_NAME_LEN]'"
    )

    addr: String = String(32)
    uid: Int32 = Int32()
    pid: Int32 = Int32()
    mod_id: Int16 = Int16()
    is_logger: Int16 = Int16()
    is_unique: Int16 = Int16()
    port: Uint16 = Uint16()
    name: String = String(32)


@pyrtma.message_def
class MDF_CLIENT_SET_NAME(MessageData, metaclass=MessageMeta):
    type_id: ClassVar[int] = 34
    type_name: ClassVar[str] = "CLIENT_SET_NAME"
    type_hash: ClassVar[int] = 0x34465AE2
    type_size: ClassVar[int] = 32
    type_source: ClassVar[str] = "core_defs.yaml"
    type_def: ClassVar[str] = (
        "'CLIENT_SET_NAME:\n  id: 34\n  fields:\n    name: char[MAX_NAME_LEN]'"
    )

    name: String = String(32)


@pyrtma.message_def
class MDF_RTMA_LOG(MessageData, metaclass=MessageMeta):
    type_id: ClassVar[int] = 40
    type_name: ClassVar[str] = "RTMA_LOG"
    type_hash: ClassVar[int] = 0x975F197F
    type_size: ClassVar[int] = 1936
    type_source: ClassVar[str] = "core_defs.yaml"
    type_def: ClassVar[str] = (
        "'RTMA_LOG:\n  id: 40\n  fields:\n    time: double\n    level: int32\n    lineno: int32\n    name: char[128]\n    pathname: char[512]\n    funcname: char[256]\n    message: char[MAX_LOG_LENGTH]'"
    )

    time: Double = Double()
    level: Int32 = Int32()
    lineno: Int32 = Int32()
    name: String = String(128)
    pathname: String = String(512)
    funcname: String = String(256)
    message: String = String(1024)


@pyrtma.message_def
class MDF_RTMA_LOG_CRITICAL(MessageData, metaclass=MessageMeta):
    type_id: ClassVar[int] = 41
    type_name: ClassVar[str] = "RTMA_LOG_CRITICAL"
    type_hash: ClassVar[int] = 0xDEBCA500
    type_size: ClassVar[int] = 1936
    type_source: ClassVar[str] = "core_defs.yaml"
    type_def: ClassVar[str] = (
        "'RTMA_LOG_CRITICAL:\n  id: 41\n  fields:\n    time: double\n    level: int32\n    lineno: int32\n    name: char[128]\n    pathname: char[512]\n    funcname: char[256]\n    message: char[MAX_LOG_LENGTH]'"
    )

    time: Double = Double()
    level: Int32 = Int32()
    lineno: Int32 = Int32()
    name: String = String(128)
    pathname: String = String(512)
    funcname: String = String(256)
    message: String = String(1024)


@pyrtma.message_def
class MDF_RTMA_LOG_ERROR(MessageData, metaclass=MessageMeta):
    type_id: ClassVar[int] = 42
    type_name: ClassVar[str] = "RTMA_LOG_ERROR"
    type_hash: ClassVar[int] = 0x45D2E328
    type_size: ClassVar[int] = 1936
    type_source: ClassVar[str] = "core_defs.yaml"
    type_def: ClassVar[str] = (
        "'RTMA_LOG_ERROR:\n  id: 42\n  fields:\n    time: double\n    level: int32\n    lineno: int32\n    name: char[128]\n    pathname: char[512]\n    funcname: char[256]\n    message: char[MAX_LOG_LENGTH]'"
    )

    time: Double = Double()
    level: Int32 = Int32()
    lineno: Int32 = Int32()
    name: String = String(128)
    pathname: String = String(512)
    funcname: String = String(256)
    message: String = String(1024)


@pyrtma.message_def
class MDF_RTMA_LOG_WARNING(MessageData, metaclass=MessageMeta):
    type_id: ClassVar[int] = 43
    type_name: ClassVar[str] = "RTMA_LOG_WARNING"
    type_hash: ClassVar[int] = 0xE6399270
    type_size: ClassVar[int] = 1936
    type_source: ClassVar[str] = "core_defs.yaml"
    type_def: ClassVar[str] = (
        "'RTMA_LOG_WARNING:\n  id: 43\n  fields:\n    time: double\n    level: int32\n    lineno: int32\n    name: char[128]\n    pathname: char[512]\n    funcname: char[256]\n    message: char[MAX_LOG_LENGTH]'"
    )

    time: Double = Double()
    level: Int32 = Int32()
    lineno: Int32 = Int32()
    name: String = String(128)
    pathname: String = String(512)
    funcname: String = String(256)
    message: String = String(1024)


@pyrtma.message_def
class MDF_RTMA_LOG_INFO(MessageData, metaclass=MessageMeta):
    type_id: ClassVar[int] = 44
    type_name: ClassVar[str] = "RTMA_LOG_INFO"
    type_hash: ClassVar[int] = 0x607B81E9
    type_size: ClassVar[int] = 1936
    type_source: ClassVar[str] = "core_defs.yaml"
    type_def: ClassVar[str] = (
        "'RTMA_LOG_INFO:\n  id: 44\n  fields:\n    time: double\n    level: int32\n    lineno: int32\n    name: char[128]\n    pathname: char[512]\n    funcname: char[256]\n    message: char[MAX_LOG_LENGTH]'"
    )

    time: Double = Double()
    level: Int32 = Int32()
    lineno: Int32 = Int32()
    name: String = String(128)
    pathname: String = String(512)
    funcname: String = String(256)
    message: String = String(1024)


@pyrtma.message_def
class MDF_RTMA_LOG_DEBUG(MessageData, metaclass=MessageMeta):
    type_id: ClassVar[int] = 45
    type_name: ClassVar[str] = "RTMA_LOG_DEBUG"
    type_hash: ClassVar[int] = 0x45EDC532
    type_size: ClassVar[int] = 1936
    type_source: ClassVar[str] = "core_defs.yaml"
    type_def: ClassVar[str] = (
        "'RTMA_LOG_DEBUG:\n  id: 45\n  fields:\n    time: double\n    level: int32\n    lineno: int32\n    name: char[128]\n    pathname: char[512]\n    funcname: char[256]\n    message: char[MAX_LOG_LENGTH]'"
    )

    time: Double = Double()
    level: Int32 = Int32()
    lineno: Int32 = Int32()
    name: String = String(128)
    pathname: String = String(512)
    funcname: String = String(256)
    message: String = String(1024)


@pyrtma.message_def
class MDF_TIMING_MESSAGE(MessageData, metaclass=MessageMeta):
    type_id: ClassVar[int] = 80
    type_name: ClassVar[str] = "TIMING_MESSAGE"
    type_hash: ClassVar[int] = 0xDAA7503D
    type_size: ClassVar[int] = 20808
    type_source: ClassVar[str] = "core_defs.yaml"
    type_def: ClassVar[str] = (
        "'TIMING_MESSAGE:\n  id: 80\n  fields:\n    timing: uint16[MAX_MESSAGE_TYPES]\n    ModulePID: int32[MAX_MODULES]\n    send_time: double'"
    )

    timing: IntArray[Uint16] = IntArray(Uint16, 10000)
    ModulePID: IntArray[Int32] = IntArray(Int32, 200)
    send_time: Double = Double()


@pyrtma.message_def
class MDF_FORCE_DISCONNECT(MessageData, metaclass=MessageMeta):
    type_id: ClassVar[int] = 82
    type_name: ClassVar[str] = "FORCE_DISCONNECT"
    type_hash: ClassVar[int] = 0x335C7BBF
    type_size: ClassVar[int] = 4
    type_source: ClassVar[str] = "core_defs.yaml"
    type_def: ClassVar[str] = (
        "'FORCE_DISCONNECT:\n  id: 82\n  fields:\n    mod_id: int32'"
    )

    mod_id: Int32 = Int32()


@pyrtma.message_def
class MDF_PAUSE_SUBSCRIPTION(MessageData, metaclass=MessageMeta):
    type_id: ClassVar[int] = 85
    type_name: ClassVar[str] = "PAUSE_SUBSCRIPTION"
    type_hash: ClassVar[int] = 0x22338A6D
    type_size: ClassVar[int] = 4
    type_source: ClassVar[str] = "core_defs.yaml"
    type_def: ClassVar[str] = (
        "'PAUSE_SUBSCRIPTION:\n  id: 85\n  fields:\n    msg_type: MSG_TYPE'"
    )

    msg_type: Int32 = Int32()


@pyrtma.message_def
class MDF_RESUME_SUBSCRIPTION(MessageData, metaclass=MessageMeta):
    type_id: ClassVar[int] = 86
    type_name: ClassVar[str] = "RESUME_SUBSCRIPTION"
    type_hash: ClassVar[int] = 0xC56A97F2
    type_size: ClassVar[int] = 4
    type_source: ClassVar[str] = "core_defs.yaml"
    type_def: ClassVar[str] = (
        "'RESUME_SUBSCRIPTION:\n  id: 86\n  fields:\n    msg_type: MSG_TYPE'"
    )

    msg_type: Int32 = Int32()


# User Context
if _ver_parse(__version__) >= _ver_parse("2.3.0"):
    _update_context(__name__)



class DataLoggerError(Exception):
    pass


class BasePathNotFound(DataLoggerError):
    pass


class MissingMetadata(DataLoggerError):
    pass


class DataCollectionNotConfigured(DataLoggerError):
    pass


class DataCollectionNotFound(DataLoggerError):
    pass


class DataCollectionInProgress(DataLoggerError):
    pass


class InvalidFormatter(DataLoggerError):
    pass


class DataSetExistsError(DataLoggerError):
    pass


class DataCollectionThreadError(DataLoggerError):
    pass


class DataFormatterKeyError(DataLoggerError):
    pass


class DataCollectionFullError(DataLoggerError):
    pass


class InvalidMetadata(DataLoggerError):
    pass



"""Run with `python -m pyrtma.unified_console_log [OPTIONS]` (use -h for help)"""

import logging
import os
from .client import Client
from . import core_defs as cd
from .client_logging import RTMALogger
from .exceptions import MessageManagerNotFound, ConnectionLost
from typing import Optional, Union, cast

RTMA_LOG_MSG = Union[
    cd.MDF_RTMA_LOG,
    cd.MDF_RTMA_LOG_DEBUG,
    cd.MDF_RTMA_LOG_INFO,
    cd.MDF_RTMA_LOG_WARNING,
    cd.MDF_RTMA_LOG_ERROR,
    cd.MDF_RTMA_LOG_CRITICAL,
]


class RTMA_UnifiedLog(object):

    def __init__(self, log_level: int = logging.INFO, server: Optional[str] = None):
        self.server = server
        self.log_level = log_level
        self._running = False

        self.mod = Client(name="UnifiedLog_Module")
        self.local_log = self.mod.logger
        self.mod.logger.enable_rtma = False
        self.mod.logger.set_all_levels(log_level)

        self.remote_log = RTMALogger(
            log_name="Remote", rtma_client=Client(), level=log_level
        )
        self.remote_log.enable_console = True
        self.remote_log.enable_rtma = False

    def init_rtma(self):
        if self.server is None:
            self.server = "127.0.0.1:7111"
        self.mod.info(f"Attempting connection to RTMA @ {self.server}")
        self.mod.connect(self.server, allow_multiple=True)

        self.mod.info("Successfully connected to RTMA")
        self.rtma_subscribe()

    def rtma_subscribe(self):
        msg_list = [cd.MT_EXIT, cd.MT_RTMA_LOG]
        if self.log_level <= logging.DEBUG:
            msg_list.append(cd.MT_RTMA_LOG_DEBUG)

        if self.log_level <= logging.INFO:
            msg_list.append(cd.MT_RTMA_LOG_INFO)

        if self.log_level <= logging.WARNING:
            msg_list.append(cd.MT_RTMA_LOG_WARNING)

        if self.log_level <= logging.ERROR:
            msg_list.append(cd.MT_RTMA_LOG_ERROR)

        if self.log_level <= logging.CRITICAL:
            msg_list.append(cd.MT_RTMA_LOG_CRITICAL)

        self.mod.subscribe(msg_list)
        self.mod.info(f"Subscribed to: {msg_list}")
        self.mod.send_module_ready()

    def disconnect(self):
        if self.mod.connected:
            self.mod.info("Disconnecting from RTMA")
            self.mod.disconnect()

    def make_log_record_dict(self, msg: RTMA_LOG_MSG):
        log_dict = {
            "name": msg.name,
            "log_name": msg.name,
            "msg": msg.message,
            "levelname": logging.getLevelName(msg.level),
            "levelno": msg.level,
            "pathname": msg.pathname,
            "filename": msg.pathname,
            "lineno": msg.lineno,
            "funcName": msg.funcname,
            "created": msg.time,
            "msecs": (msg.time - int(msg.time)) * 1000,
            "relativeCreated": (msg.time - logging._startTime) * 1000,  # type: ignore
        }
        try:
            log_dict["filename"] = os.path.basename(msg.pathname)
            log_dict["module"] = os.path.splitext(log_dict["filename"])[0]
        except (TypeError, ValueError, AttributeError):
            log_dict["filename"] = msg.pathname
            log_dict["module"] = "Unknown module"

        return log_dict

    def process_log_msg(self, msg: RTMA_LOG_MSG):
        if msg.level < self.log_level:
            return
        log_dict = self.make_log_record_dict(msg)
        log_record = logging.makeLogRecord(log_dict)
        if log_record.levelno >= logging.WARNING:
            print("\a", flush=True)
        if self.remote_log.console_handler:
            self.remote_log.console_handler.emit(log_record)

    def start(self):

        log_msg_ids = (
            cd.MT_RTMA_LOG,
            cd.MT_RTMA_LOG_DEBUG,
            cd.MT_RTMA_LOG_INFO,
            cd.MT_RTMA_LOG_WARNING,
            cd.MT_RTMA_LOG_ERROR,
            cd.MT_RTMA_LOG_CRITICAL,
        )
        if self.server is None:
            self.server = "127.0.0.1:7111"
        self._running = True
        try:
            self.mod.info("Starting main loop")
            while self._running:
                try:
                    self.init_rtma()
                    while self._running:
                        msg = self.mod.read_message(1)
                        if msg:
                            if msg.type_id == cd.MT_EXIT:
                                self.mod.info("Received EXIT signal")
                                self._running = False
                            elif msg.type_id in log_msg_ids:
                                msg.data = cast(RTMA_LOG_MSG, msg.data)
                                self.process_log_msg(msg.data)
                except MessageManagerNotFound:
                    continue
                except ConnectionLost:
                    self.mod.error("RTMA Connection lost, attempting to reconnect")
        except KeyboardInterrupt:
            self.mod.info("Keyboard Interrupt")
        finally:
            self.mod.disconnect()
            self.mod.info("Exiting")


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-a",
        "--addr",
        dest="addr",
        type=str,
        default="127.0.0.1",
        help="MessageManager server IP address as a string. Default is 127.0.0.1",
    )
    parser.add_argument(
        "-p",
        "--port",
        dest="port",
        type=int,
        default=7111,
        help="Listener port. Default is 7111.",
    )
    parser.add_argument(
        "-l",
        "--log-level",
        dest="log_level",
        choices=["DEBUG", "INFO", "WARN", "ERROR", "CRITICAL"],
        default="INFO",
        help="Logging Level",
    )

    args = parser.parse_args()

    server_name = f"{args.addr}:{args.port}"

    log_level = logging.getLevelNamesMapping().get(args.log_level) or logging.INFO

    RL = RTMA_UnifiedLog(log_level=log_level, server=server_name)
    RL.start()
